import json
from pathlib import Path
from types import SimpleNamespace

import pytest

from biblicus.evaluation import benchmark_runner
from biblicus.evaluation.metrics import entity_metrics
from biblicus.evaluation.ocr_benchmark import OCRBenchmark
from biblicus.graph.extractors import dependency_relations, ner_entities, simple_entities
from biblicus.migration import _migrate_raw_items, _migrate_snapshots
from biblicus.sync.amplify_publisher import AmplifyPublisher


def test_benchmark_config_load_parses_categories(tmp_path):
    config_path = tmp_path / "bench.yaml"
    config_path.write_text(
        """
benchmark_name: demo
categories:
  forms:
    dataset: demo_ds
pipelines:
  - pipeline1.yml
""",
        encoding="utf-8",
    )
    config = benchmark_runner.BenchmarkConfig.load(config_path)
    assert "forms" in config.categories
    assert config.categories["forms"].dataset == "demo_ds"
    assert config.pipelines[0].name == "pipeline1.yml"


class DummyReport:
    avg_f1 = 0.9
    avg_recall = 0.8
    avg_precision = 0.85
    avg_word_error_rate = 0.1
    avg_lcs_ratio = 0.95
    avg_bigram_overlap = 0.7
    avg_sequence_accuracy = 0.88
    total_documents = 1


class DummyBenchmark:
    def __init__(self):
        self.calls = []

    def run_pipeline(self, pipeline_path):
        self.calls.append(Path(pipeline_path).name)
        return DummyReport()

    def evaluate_extraction(self, snapshot_reference, ground_truth_dir=None):
        return DummyReport()


def test_evaluate_category_tracks_best_pipeline(tmp_path, monkeypatch):
    category = benchmark_runner.CategoryConfig(
        name="forms",
        dataset="demo",
        corpus_path=tmp_path,
        ground_truth_subdir="gt",
        primary_metric="f1",
    )

    class DummyCorpus:
        def __init__(self, root: Path):
            self.meta_dir = root / "metadata"
            self.meta_dir.mkdir(parents=True, exist_ok=True)

        def extract(self, extractor_id: str, config: dict):  # noqa: ARG002
            return SimpleNamespace(snapshot_id="snap1")

        def extraction_snapshot_dir(self, extractor_id: str, snapshot_id: str) -> Path:  # noqa: ARG002
            return tmp_path / "snap"

    dummy_corpus = DummyCorpus(tmp_path)
    gt_dir = dummy_corpus.meta_dir / "gt"
    gt_dir.mkdir(parents=True, exist_ok=True)
    snap_dir = dummy_corpus.extraction_snapshot_dir("pipeline", "snap1")
    (snap_dir / "text").mkdir(parents=True, exist_ok=True)
    (snap_dir / "text" / "doc.txt").write_text("hello", encoding="utf-8")
    (gt_dir / "doc.txt").write_text("hello", encoding="utf-8")

    pipeline_path = tmp_path / "pipe.yml"
    pipeline_path.write_text("extractor_id: pipeline\nconfig: {}\n", encoding="utf-8")

    runner = benchmark_runner.BenchmarkRunner(
        benchmark_runner.BenchmarkConfig(
            benchmark_name="demo",
            categories={"forms": category},
            pipelines=[pipeline_path],
            aggregate_weights={},
        )
    )
    monkeypatch.setattr(benchmark_runner, "Corpus", SimpleNamespace(open=lambda p: dummy_corpus))
    monkeypatch.setattr(benchmark_runner, "OCRBenchmark", lambda c: DummyBenchmark())
    result = runner.run_category(category)
    assert result.best_pipeline == "pipe"
    assert result.best_score > 0


@pytest.mark.parametrize(
    ("value", "entity_type", "expected"),
    [
        ("$1,234.50", "total", "1234.50"),
        ("Example LLC", "company", "example"),
        ("123 st. ave", "address", "123 street. avenue"),
        ("Date: 2024/05/01", "date", "2024-05-01"),
    ],
)
def test_entity_normalization_branches(value, entity_type, expected):
    assert entity_metrics.normalize_entity_value(value, entity_type) == expected


def test_entity_metrics_overall_counts():
    results = entity_metrics.calculate_entity_metrics(
        ground_truth={"date": "2024-01-01", "total": "$10.00"},
        extracted={"date": "2024-01-01", "total": "10"},
        entity_types=["date", "total"],
        fuzzy_threshold=0.5,
    )
    assert results["overall"]["exact_matches"] == 1.0
    assert results["overall"]["fuzzy_matches"] == 1.0


def test_ocr_benchmark_handles_missing_catalog_item(tmp_path):
    class DummyCorpus:
        def __init__(self, root: Path):
            self.root = root
            self.meta_dir = root / "metadata"
            self.meta_dir.mkdir(parents=True, exist_ok=True)

        def extraction_snapshot_dir(self, extractor_id: str, snapshot_id: str) -> Path:  # noqa: ARG002
            return self.root / "snap"

        def get_item(self, item_id):  # noqa: ARG002
            raise KeyError("missing")

    corpus = DummyCorpus(tmp_path)
    snap_dir = corpus.extraction_snapshot_dir("pipeline", "snap1")
    (snap_dir / "text").mkdir(parents=True, exist_ok=True)
    (snap_dir / "text" / "doc1.txt").write_text("hello", encoding="utf-8")
    gt_dir = corpus.meta_dir / "gt"
    gt_dir.mkdir(parents=True, exist_ok=True)
    (gt_dir / "doc1.txt").write_text("hello", encoding="utf-8")

    benchmark = OCRBenchmark(corpus)
    report = benchmark.evaluate_extraction("snap1", ground_truth_dir=gt_dir)
    assert report.total_documents == 1


def test_ocr_benchmark_uses_catalog_relpath(tmp_path):
    class DummyCorpus:
        def __init__(self, root: Path):
            self.root = root
            self.meta_dir = root / "metadata"
            self.meta_dir.mkdir(parents=True, exist_ok=True)

        def extraction_snapshot_dir(self, extractor_id: str, snapshot_id: str) -> Path:  # noqa: ARG002
            return self.root / "snap"

        def get_item(self, item_id):
            return SimpleNamespace(relpath=f"raw/{item_id}.png")

    corpus = DummyCorpus(tmp_path)
    snap_dir = corpus.extraction_snapshot_dir("pipeline", "snap1")
    (snap_dir / "text").mkdir(parents=True, exist_ok=True)
    (snap_dir / "text" / "doc1.txt").write_text("hello", encoding="utf-8")
    gt_dir = corpus.meta_dir / "gt"
    gt_dir.mkdir(parents=True, exist_ok=True)
    (gt_dir / "doc1.txt").write_text("hello", encoding="utf-8")

    benchmark = OCRBenchmark(corpus)
    report = benchmark.evaluate_extraction("snap1", ground_truth_dir=gt_dir)
    assert report.total_documents == 1


def test_dependency_relations_validates_config_dict(monkeypatch):
    extractor = dependency_relations.DependencyRelationsGraphExtractor()
    monkeypatch.setattr(dependency_relations, "_extract_entities", lambda **kwargs: [("a", "LABEL")])
    dummy_doc = type("Doc", (), {"__iter__": lambda self: iter([])})
    monkeypatch.setattr(dependency_relations, "_load_doc", lambda text, model_name: dummy_doc())
    corpus = SimpleNamespace()
    item = SimpleNamespace(id="item1", title="t", relpath="r")
    extraction = extractor.extract_graph(
        corpus=corpus,
        item=item,
        extracted_text="abc",
        config={"model": "en", "min_entity_length": 1},
    )
    assert extraction.nodes


def test_ner_entities_validates_config_dict(monkeypatch):
    extractor = ner_entities.NerEntitiesGraphExtractor()
    monkeypatch.setattr(ner_entities, "_extract_entities", lambda **kwargs: [("Name", "PERSON")])
    extraction = extractor.extract_graph(
        corpus=SimpleNamespace(),
        item=SimpleNamespace(id="i1", title="t", relpath="r"),
        extracted_text="John went to Paris",
        config={"model": "en", "min_entity_length": 1},
    )
    assert extraction.nodes is not None


def test_simple_entities_validates_config_dict():
    extractor = simple_entities.SimpleEntitiesGraphExtractor()
    extraction = extractor.extract_graph(
        corpus=SimpleNamespace(),
        item=SimpleNamespace(id="i1", title="t", relpath="r"),
        extracted_text="Token test",
        config={"min_entity_length": 1, "include_item_node": False, "max_entity_words": 3},
    )
    assert extraction.nodes


def test_migration_functions_cleanup(tmp_path):
    raw_dir = tmp_path / "raw"
    raw_dir.mkdir()
    (raw_dir / "file.txt").write_text("data", encoding="utf-8")
    stats = {"moved_raw_items": 0, "moved_extraction_snapshots": 0, "moved_graph_snapshots": 0,
             "moved_analysis_runs": 0, "moved_retrieval_snapshots": 0}
    _migrate_raw_items(root=tmp_path, force=True, stats=stats)
    assert stats["moved_raw_items"] == 1

    snapshots_root = tmp_path / "metadata" / "snapshots" / "analysis"
    snapshots_root.mkdir(parents=True)
    (snapshots_root / "dummy").mkdir()
    _migrate_snapshots(root=tmp_path, meta_dir=tmp_path / "metadata", force=True, stats=stats)
    assert stats["moved_analysis_runs"] >= 0


def test_amplify_publisher_reads_config_file(tmp_path, monkeypatch):
    monkeypatch.setattr(Path, "home", lambda: tmp_path)
    env_path = tmp_path / ".biblicus"
    env_path.mkdir(exist_ok=True)
    amplify_env = env_path / "amplify.env"
    amplify_env.write_text(
        "AMPLIFY_APPSYNC_ENDPOINT=https://example\nAMPLIFY_API_KEY=key\nAMPLIFY_S3_BUCKET=bkt",
        encoding="utf-8",
    )
    monkeypatch.delenv("AMPLIFY_APPSYNC_ENDPOINT", raising=False)
    monkeypatch.delenv("AMPLIFY_API_KEY", raising=False)
    monkeypatch.delenv("AMPLIFY_S3_BUCKET", raising=False)

    publisher = AmplifyPublisher(corpus_name="demo")
    assert publisher.appsync_endpoint == "https://example"
    assert publisher.api_key == "key"
    assert publisher.s3_bucket == "bkt"


def test_amplify_publisher_retries_network_error(monkeypatch):
    monkeypatch.setattr(Path, "home", lambda: Path.cwd())
    monkeypatch.setenv("AMPLIFY_APPSYNC_ENDPOINT", "https://example")
    monkeypatch.setenv("AMPLIFY_API_KEY", "key")
    monkeypatch.setenv("AMPLIFY_S3_BUCKET", "bucket")
    publisher = AmplifyPublisher(corpus_name="demo")
    attempts = {"count": 0}

    def fake_exec(query, variables):  # noqa: ARG002
        attempts["count"] += 1
        if attempts["count"] < 2:
            raise Exception("Network glitch")
        return {}

    monkeypatch.setattr(publisher, "_execute_graphql", fake_exec)
    publisher._create_catalog_item(
        SimpleNamespace(
            id="1",
            relpath="a",
            sha256="x",
            bytes=1,
            media_type="text/plain",
            title=None,
            tags=[],
            metadata={},
            source_uri=None,
        )
    )
    assert attempts["count"] == 2



from pathlib import Path
from types import SimpleNamespace

from biblicus.corpus import Corpus
from biblicus.models import CorpusCatalog


def _make_corpus(tmp_path: Path) -> Corpus:
    corpus = Corpus(tmp_path)
    corpus.meta_dir.mkdir(parents=True, exist_ok=True)
    catalog = CorpusCatalog(
        schema_version=2,
        generated_at="2024-01-01T00:00:00Z",
        corpus_uri=tmp_path.as_uri(),
        raw_dir="raw",
        items={},
        order=[],
    )
    corpus._write_catalog(catalog)
    return corpus


def test_resolve_remote_source_name_defaults():
    corpus = Corpus(Path("/tmp"))
    source = SimpleNamespace(kind="s3", bucket="my-bucket", container=None, name=None)
    assert corpus._resolve_remote_source_name(source) == "my-bucket"
    source = SimpleNamespace(kind="azure-blob", bucket=None, container="my-container", name=None)
    assert corpus._resolve_remote_source_name(source) == "my-container"


def test_relative_remote_key_strips_prefix():
    corpus = Corpus(Path("/tmp"))
    key = "prefix/path/file.txt"
    assert corpus._relative_remote_key(key, prefix="prefix/") == "path/file.txt"


def test_remote_item_unchanged_checks_etag_and_last_modified():
    corpus = Corpus(Path("/tmp"))
    existing = SimpleNamespace(metadata={"biblicus": {"source_etag": "etag1", "source_last_modified": "y"}})
    item = SimpleNamespace(etag="etag1", last_modified="z")
    assert corpus._remote_item_unchanged(existing, item)
    item2 = SimpleNamespace(etag=None, last_modified="y")
    assert corpus._remote_item_unchanged(existing, item2)



"""
Agentic text slicing using virtual file edits.
"""

from __future__ import annotations

import re
from typing import List

from .models import TextSliceRequest, TextSliceResult, TextSliceSegment
from .tool_loop import request_confirmation, run_tool_loop

_SLICE_MARKER = "<slice/>"


def apply_text_slice(request: TextSliceRequest) -> TextSliceResult:
    """
    Apply text slicing using a language model.

    :param request: Text slice request.
    :type request: TextSliceRequest
    :return: Text slice result.
    :rtype: TextSliceResult
    :raises ValueError: If model output is invalid or text is modified. Empty outputs trigger
        a confirmation round and return a warning when confirmed.
    """
    if request.mock_marked_up_text is not None:
        return _build_mock_result(request, request.mock_marked_up_text)

    warnings: List[str] = []
    result = run_tool_loop(
        text=request.text,
        client=request.client,
        system_prompt=request.system_prompt,
        prompt_template=request.prompt_template,
        max_rounds=request.max_rounds,
        max_edits_per_round=request.max_edits_per_round,
        apply_str_replace=_apply_slice_replace,
    )
    if not result.done:
        if result.last_error:
            raise ValueError(f"Text slice failed: {result.last_error}")
        warnings.append("Text slice reached max rounds without done=true")
    if result.text == request.text:
        if result.last_error:
            raise ValueError(result.last_error)
        confirmation = request_confirmation(
            result=result,
            text=result.text,
            client=request.client,
            system_prompt=request.system_prompt,
            prompt_template=request.prompt_template,
            max_rounds=2,
            max_edits_per_round=request.max_edits_per_round,
            apply_str_replace=_apply_slice_replace,
            confirmation_message=_build_empty_confirmation_message(result.text),
        )
        if not confirmation.done:
            if confirmation.last_error:
                raise ValueError(f"Text slice failed: {confirmation.last_error}")
            warnings.append("Text slice confirmation reached max rounds without done=true")
        _validate_preserved_text(original=request.text, marked_up=confirmation.text)
        slices = _extract_slices(marked_up_text=confirmation.text)
        if confirmation.text == request.text:
            warnings.append("Text slice returned no markers; model confirmed single slice")
        if not slices:
            raise ValueError("Text slice produced no slices")
        return TextSliceResult(
            marked_up_text=confirmation.text,
            slices=slices,
            warnings=warnings,
        )
    _validate_preserved_text(original=request.text, marked_up=result.text)
    slices = _extract_slices(marked_up_text=result.text)
    if not slices:
        raise ValueError("Text slice produced no slices")
    return TextSliceResult(marked_up_text=result.text, slices=slices, warnings=warnings)


def _build_mock_result(request: TextSliceRequest, marked_up_text: str) -> TextSliceResult:
    if marked_up_text == request.text:
        raise ValueError("Text slice produced no markers")
    _validate_preserved_text(original=request.text, marked_up=marked_up_text)
    slices = _extract_slices(marked_up_text=marked_up_text)
    return TextSliceResult(marked_up_text=marked_up_text, slices=slices, warnings=[])


def _apply_slice_replace(text: str, old_str: str, new_str: str) -> str:
    occurrences = text.count(old_str)
    if occurrences == 0:
        raise ValueError("Text slice replacement old_str not found")
    if occurrences > 1:
        raise ValueError("Text slice replacement old_str is not unique")
    _validate_replace_text(old_str, new_str)
    return text.replace(old_str, new_str, 1)


def _validate_replace_text(old_str: str, new_str: str) -> None:
    if _strip_slice_markers(old_str) != _strip_slice_markers(new_str):
        raise ValueError("Text slice replacements may only insert slice markers")


def _validate_preserved_text(*, original: str, marked_up: str) -> None:
    if _strip_slice_markers(marked_up) != original:
        raise ValueError("Text slice edits modified the source text")


def _strip_slice_markers(text: str) -> str:
    return text.replace(_SLICE_MARKER, "")


def _extract_slices(*, marked_up_text: str) -> List[TextSliceSegment]:
    marker_pattern = re.compile(re.escape(_SLICE_MARKER))
    slices: List[TextSliceSegment] = []
    cursor = 0
    original_index = 0

    for match in marker_pattern.finditer(marked_up_text):
        chunk = marked_up_text[cursor : match.start()]
        if chunk:
            slice_end = original_index + len(chunk)
            slices.append(
                TextSliceSegment(
                    index=len(slices) + 1,
                    start_char=original_index,
                    end_char=slice_end,
                    text=chunk,
                )
            )
            original_index = slice_end
        cursor = match.end()

    tail = marked_up_text[cursor:]
    if tail:
        slice_end = original_index + len(tail)
        slices.append(
            TextSliceSegment(
                index=len(slices) + 1,
                start_char=original_index,
                end_char=slice_end,
                text=tail,
            )
        )

    return slices


def _build_empty_confirmation_message(text: str) -> str:
    return (
        "No slice markers were inserted. If the text should remain a single slice, "
        "call done again without changes. Otherwise insert <slice/> markers at the "
        "boundaries of the requested slices.\n"
        "Current text:\n"
        f"---\n{text}\n---"
    )



from types import SimpleNamespace

from biblicus.analysis import markov


def test_markov_llm_observation_parallel(monkeypatch):
    monkeypatch.setattr(
        markov,
        "generate_completion",
        lambda client, system_prompt, user_prompt: '{"label": "p", "label_confidence": 0.9, "summary": "s"}',
    )
    monkeypatch.setattr(markov, "_parse_json_object", lambda text, error_label: {"label": "p", "label_confidence": 0.9, "summary": "s"})
    # speed up timing
    monkeypatch.setattr(markov.time, "perf_counter", lambda: 0.0)
    config = SimpleNamespace(
        llm_observations=SimpleNamespace(
            enabled=True,
            client=SimpleNamespace(response_format=None),
            prompt_template="{segment}",
            system_prompt=None,
            cache=SimpleNamespace(enabled=False),
            max_workers=2,
        ),
        embeddings=SimpleNamespace(enabled=False),
    )
    segments = [
        markov.MarkovAnalysisSegment(item_id="i", segment_index=1, text="body1"),
        markov.MarkovAnalysisSegment(item_id="i", segment_index=2, text="body2"),
    ]
    observations = markov._build_observations(segments=segments, config=config, cache_context=None)
    assert all(obs.llm_label == "p" for obs in observations)



#!/usr/bin/env python3
"""
Sync an existing corpus to Amplify.

Usage:
    python scripts/sync_corpus_to_amplify.py corpora/markov-demo
"""

import os
import sys
from pathlib import Path
from datetime import datetime

# Add src to path
sys.path.insert(0, str(Path(__file__).parent.parent / 'src'))

# Load Amplify config
config_path = Path.home() / '.biblicus' / 'amplify.env'
if config_path.exists():
    with open(config_path) as f:
        for line in f:
            if '=' in line:
                key, value = line.strip().split('=', 1)
                os.environ[key] = value

from biblicus.sync.amplify_publisher import AmplifyPublisher
from biblicus.corpus import Corpus

def main():
    if len(sys.argv) < 2:
        print('Usage: python scripts/sync_corpus_to_amplify.py <corpus-path>')
        return 1

    corpus_path = Path(sys.argv[1])
    if not corpus_path.exists():
        print(f'Error: Corpus path not found: {corpus_path}')
        return 1

    print(f'Syncing corpus: {corpus_path}')

    # Load the corpus
    corpus = Corpus(corpus_path)
    corpus_name = corpus_path.name

    print(f'Corpus name: {corpus_name}')

    # Create publisher
    publisher = AmplifyPublisher(corpus_name)

    # Create or update Corpus record
    mutation = '''
    mutation CreateCorpus($input: CreateCorpusInput!) {
      createCorpus(input: $input) {
        id
        name
        s3Prefix
        status
        createdAt
      }
    }
    '''

    variables = {
        'input': {
            'name': corpus_name,
            's3Prefix': f'{corpus_name}/',
            'status': 'IDLE',
            'createdAt': datetime.now().isoformat() + 'Z',
            'updatedAt': datetime.now().isoformat() + 'Z'
        }
    }

    try:
        result = publisher._execute_graphql(mutation, variables)
        print(f'✓ Created corpus record: {corpus_name}')
    except Exception as e:
        if 'already exists' in str(e) or 'DuplicateKey' in str(e):
            print(f'✓ Corpus record already exists: {corpus_name}')
        else:
            raise

    # Sync catalog if it exists
    catalog_path = corpus.root / 'catalog.json'
    if catalog_path.exists():
        print(f'Syncing catalog from {catalog_path}...')
        result = publisher.sync_catalog(catalog_path, force=True)
        print(f'✓ Synced catalog: {result.created} created, {result.updated} updated')
    else:
        print('No catalog.json found, skipping catalog sync')

    print(f'\n✓ Corpus synced successfully!')
    print(f'View at http://localhost:5173')

    return 0

if __name__ == '__main__':
    sys.exit(main())



"""
Tesseract OCR extractor plugin.

This extractor performs optical character recognition on image items using Tesseract OCR.
It supports layout-aware OCR by reading region metadata from previous pipeline stages.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field

from ..corpus import Corpus
from ..errors import ExtractionSnapshotFatalError
from ..models import CatalogItem, ExtractedText, ExtractionStageOutput
from .base import TextExtractor


class TesseractExtractorConfig(BaseModel):
    """
    Configuration for the Tesseract OCR extractor.

    :ivar min_confidence: Minimum per-word confidence to include in output (0.0-1.0).
    :vartype min_confidence: float
    :ivar joiner: Joiner used to combine recognized text segments.
    :vartype joiner: str
    :ivar lang: Tesseract language code (e.g., 'eng', 'fra', 'chi_sim').
    :vartype lang: str
    :ivar psm: Page segmentation mode (0-13). Default 3 is fully automatic.
    :vartype psm: int
    :ivar oem: OCR Engine Mode (0=Legacy, 1=LSTM, 2=Combined, 3=Default).
    :vartype oem: int
    :ivar use_layout_metadata: Read region metadata from previous pipeline stage.
    :vartype use_layout_metadata: bool
    """

    model_config = ConfigDict(extra="forbid")

    min_confidence: float = Field(default=0.0, ge=0.0, le=1.0)
    joiner: str = Field(default="\n")
    lang: str = Field(default="eng")
    psm: int = Field(default=3, ge=0, le=13)
    oem: int = Field(default=3, ge=0, le=3)
    use_layout_metadata: bool = Field(default=False)


class TesseractExtractor(TextExtractor):
    """
    Extractor plugin that performs optical character recognition using Tesseract OCR.

    This extractor can operate in two modes:
    1. Standard mode: OCR the entire image
    2. Layout-aware mode: Read region metadata from previous stage and OCR each region separately

    :ivar extractor_id: Extractor identifier.
    :vartype extractor_id: str
    """

    extractor_id = "ocr-tesseract"

    def validate_config(self, config: Dict[str, Any]) -> BaseModel:
        """
        Validate extractor configuration and ensure Tesseract is available.

        :param config: Configuration mapping.
        :type config: dict[str, Any]
        :return: Parsed configuration model.
        :rtype: TesseractExtractorConfig
        :raises ExtractionSnapshotFatalError: If pytesseract or tesseract is missing.
        """
        try:
            import pytesseract  # noqa: F401
        except ImportError as import_error:
            raise ExtractionSnapshotFatalError(
                "Tesseract extractor requires pytesseract. "
                'Install it with pip install "biblicus[tesseract]".'
            ) from import_error

        # Check if tesseract binary is available
        try:
            import pytesseract

            pytesseract.get_tesseract_version()
        except Exception as check_error:
            raise ExtractionSnapshotFatalError(
                "Tesseract OCR binary not found. "
                "Install tesseract-ocr: brew install tesseract (macOS) or "
                "apt-get install tesseract-ocr (Ubuntu)."
            ) from check_error

        return TesseractExtractorConfig.model_validate(config)

    def extract_text(
        self,
        *,
        corpus: Corpus,
        item: CatalogItem,
        config: BaseModel,
        previous_extractions: List[ExtractionStageOutput],
    ) -> Optional[ExtractedText]:
        """
        Extract text from an image item using Tesseract OCR.

        :param corpus: Corpus containing the item bytes.
        :type corpus: Corpus
        :param item: Catalog item to extract text from.
        :type item: CatalogItem
        :param config: Parsed configuration model.
        :type config: TesseractExtractorConfig
        :param previous_extractions: Prior stage outputs for this item within the pipeline.
        :type previous_extractions: list[ExtractionStageOutput]
        :return: Extracted text payload or None when item is not an image.
        :rtype: ExtractedText or None
        """
        parsed_config = (
            config
            if isinstance(config, TesseractExtractorConfig)
            else TesseractExtractorConfig.model_validate(config)
        )

        # Skip non-image items
        if not item.media_type.startswith("image/"):
            return None

        # Get image path
        source_path = corpus.root / item.relpath

        # Check for layout metadata if layout-aware mode is enabled
        if parsed_config.use_layout_metadata and previous_extractions:
            layout_metadata = previous_extractions[-1].metadata
            if layout_metadata and "regions" in layout_metadata:
                return self._extract_with_layout(source_path, layout_metadata, parsed_config)

        # Standard mode: OCR entire image
        return self._extract_full_image(source_path, parsed_config)

    def _extract_full_image(self, image_path, config: TesseractExtractorConfig) -> ExtractedText:
        """
        Extract text from entire image using Tesseract.

        :param image_path: Path to image file.
        :type image_path: Path
        :param config: Extractor configuration.
        :type config: TesseractExtractorConfig
        :return: Extracted text with confidence.
        :rtype: ExtractedText
        """
        import pytesseract
        from PIL import Image

        # Build Tesseract configuration string
        tesseract_config = f"--oem {config.oem} --psm {config.psm}"

        # Load image
        image = Image.open(str(image_path))

        # Get detailed OCR data with confidence scores
        data = pytesseract.image_to_data(
            image,
            lang=config.lang,
            config=tesseract_config,
            output_type=pytesseract.Output.DICT,
        )

        # Extract words with confidence filtering
        words = []
        confidences = []

        for i, word_text in enumerate(data["text"]):
            if not word_text.strip():
                continue

            # pytesseract returns confidence as int 0-100, convert to float 0.0-1.0
            confidence = float(data["conf"][i]) / 100.0

            if confidence >= config.min_confidence:
                words.append(word_text)
                confidences.append(confidence)

        # Join words and calculate average confidence
        text = config.joiner.join(words) if words else ""
        avg_confidence = sum(confidences) / len(confidences) if confidences else None

        return ExtractedText(
            text=text,
            producer_extractor_id=self.extractor_id,
            confidence=avg_confidence,
        )

    def _extract_with_layout(
        self, image_path, layout_metadata: Dict[str, Any], config: TesseractExtractorConfig
    ) -> ExtractedText:
        """
        Extract text from image regions using layout metadata.

        :param image_path: Path to image file.
        :type image_path: Path
        :param layout_metadata: Layout regions and reading order from previous stage.
        :type layout_metadata: dict
        :param config: Extractor configuration.
        :type config: TesseractExtractorConfig
        :return: Extracted text with confidence.
        :rtype: ExtractedText
        """
        import pytesseract
        from PIL import Image

        # Build Tesseract configuration string
        tesseract_config = f"--oem {config.oem} --psm {config.psm}"

        # Load full image
        image = Image.open(str(image_path))

        # Get regions sorted by reading order
        regions = sorted(layout_metadata.get("regions", []), key=lambda r: r.get("order", 999))

        # OCR each region
        region_texts = []
        all_confidences = []

        for region in regions:
            bbox = region.get("bbox")
            if not bbox or len(bbox) != 4:
                continue

            # Crop region (bbox format: [x, y, w, h])
            x, y, w, h = bbox
            region_image = image.crop((x, y, x + w, y + h))

            # OCR this region
            data = pytesseract.image_to_data(
                region_image,
                lang=config.lang,
                config=tesseract_config,
                output_type=pytesseract.Output.DICT,
            )

            # Extract words from this region
            region_words = []
            for i, word_text in enumerate(data["text"]):
                if not word_text.strip():
                    continue

                confidence = float(data["conf"][i]) / 100.0
                if confidence >= config.min_confidence:
                    region_words.append(word_text)
                    all_confidences.append(confidence)

            if region_words:
                region_texts.append(config.joiner.join(region_words))

        # Join all region texts
        text = "\n\n".join(region_texts) if region_texts else ""
        avg_confidence = sum(all_confidences) / len(all_confidences) if all_confidences else None

        # Include metadata about regions processed
        metadata = {
            "regions_processed": len(regions),
            "regions_with_text": len(region_texts),
        }

        return ExtractedText(
            text=text,
            producer_extractor_id=self.extractor_id,
            confidence=avg_confidence,
            metadata=metadata,
        )



import pytest
from types import SimpleNamespace

from biblicus.analysis import topic_modeling
from biblicus.analysis.models import (
    TopicModelingConfiguration,
    TopicModelingLlmExtractionConfig,
    TopicModelingLlmExtractionMethod,
    TopicModelingLexicalProcessingConfig,
)
from biblicus.analysis.topic_modeling import TopicModelingDocument, _apply_lexical_processing


def test_topic_modeling_llm_extraction_handles_empty_response(monkeypatch):
    documents = [TopicModelingDocument(document_id="d1", source_item_id="i1", text="text")]
    config = TopicModelingLlmExtractionConfig.model_construct(
        enabled=True,
        method=TopicModelingLlmExtractionMethod.ITEMIZE,
        client={"provider": "mock", "model": "demo"},
        prompt_template="{text}",
        system_prompt=None,
        max_keywords=2,
        max_documents=1,
    )
    monkeypatch.setattr(topic_modeling, "generate_completion", lambda **kwargs: "[]")
    with pytest.raises(ValueError):
        topic_modeling._apply_llm_extraction(documents=documents, config=config)


def test_topic_modeling_lexical_processing_skips_when_disabled():
    docs = [TopicModelingDocument(document_id="d", source_item_id="", text="x")]
    cfg = TopicModelingLexicalProcessingConfig(enabled=False)
    report, processed = _apply_lexical_processing(documents=docs, config=cfg)
    assert report.status.name == "SKIPPED"
    assert processed == docs



#!/usr/bin/env python3
"""
Create test corpus data in Amplify for local testing.

Usage:
    python scripts/create_test_corpus.py
"""

import os
import sys
from pathlib import Path
from datetime import datetime

# Add src to path
sys.path.insert(0, str(Path(__file__).parent.parent / 'src'))

# Load Amplify config
config_path = Path.home() / '.biblicus' / 'amplify.env'
if config_path.exists():
    with open(config_path) as f:
        for line in f:
            if '=' in line:
                key, value = line.strip().split('=', 1)
                os.environ[key] = value

from biblicus.sync.amplify_publisher import AmplifyPublisher

def main():
    print('Creating test corpus data...')

    publisher = AmplifyPublisher('test-corpus')

    # Create a test Corpus record
    mutation = '''
    mutation CreateCorpus($input: CreateCorpusInput!) {
      createCorpus(input: $input) {
        id
        name
        s3Prefix
        status
      }
    }
    '''

    variables = {
        'input': {
            'name': 'test-corpus',
            's3Prefix': 'test-corpus/',
            'status': 'IDLE',
            'createdAt': datetime.utcnow().isoformat() + 'Z',
            'updatedAt': datetime.utcnow().isoformat() + 'Z'
        }
    }

    try:
        result = publisher._execute_graphql(mutation, variables)
        print(f'✓ Created corpus: {result["createCorpus"]["name"]}')

        # Create a test Snapshot
        snapshot_mutation = '''
        mutation CreateSnapshot($input: CreateSnapshotInput!) {
          createSnapshot(input: $input) {
            corpusId
            snapshotId
            type
            status
          }
        }
        '''

        snapshot_vars = {
            'input': {
                'corpusId': 'test-corpus',
                'snapshotId': 'snapshot-001',
                'type': 'EXTRACTION',
                'status': 'COMPLETED',
                'totalItems': 100,
                'completedItems': 100,
                'startTime': datetime.utcnow().isoformat() + 'Z'
            }
        }

        result = publisher._execute_graphql(snapshot_mutation, snapshot_vars)
        print(f'✓ Created snapshot: {result["createSnapshot"]["snapshotId"]}')

        # Create some test CatalogItems
        for i in range(3):
            item_mutation = '''
            mutation CreateCatalogItem($input: CreateCatalogItemInput!) {
              createCatalogItem(input: $input) {
                corpusId
                itemId
                relpath
              }
            }
            '''

            item_vars = {
                'input': {
                    'corpusId': 'test-corpus',
                    'itemId': f'item-{i+1:03d}',
                    'relpath': f'documents/test-doc-{i+1}.txt',
                    'sha256': 'a' * 64,
                    'bytes': 1024 * (i + 1),
                    'mediaType': 'text/plain',
                    'title': f'Test Document {i+1}',
                    'createdAt': datetime.utcnow().isoformat() + 'Z',
                    'hasExtraction': i % 2 == 0
                }
            }

            result = publisher._execute_graphql(item_mutation, item_vars)
            print(f'✓ Created catalog item: {result["createCatalogItem"]["itemId"]}')

        print('\n✓ Test data created successfully!')
        print('Open http://localhost:5173 to view the dashboard')

        return 0
    except Exception as e:
        print(f'\n✗ Failed to create test data: {e}')
        return 1

if __name__ == '__main__':
    sys.exit(main())



from __future__ import annotations

import json
import sys
import types
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional

from behave import given, then, when

from biblicus.corpus import Corpus
from features.environment import run_biblicus


@dataclass
class _FakeHmmlearnBehavior:
    predicted_states: List[int]
    include_transmat: bool = True
    include_startprob: bool = True
    startprob_values: Optional[List[float]] = None


def _ensure_fake_hmmlearn_behavior(context) -> _FakeHmmlearnBehavior:
    behavior = getattr(context, "fake_hmmlearn_behavior", None)
    if behavior is None:
        behavior = _FakeHmmlearnBehavior(predicted_states=[])
        context.fake_hmmlearn_behavior = behavior
    return behavior


def _install_fake_hmmlearn_module(context) -> None:
    already_installed = getattr(context, "_fake_hmmlearn_installed", False)
    if already_installed:
        return

    original_modules: Dict[str, object] = {}
    for name in ["hmmlearn", "hmmlearn.hmm"]:
        if name in sys.modules:
            original_modules[name] = sys.modules[name]

    behavior = _ensure_fake_hmmlearn_behavior(context)

    class GaussianHMM:  # noqa: N801 - external dependency uses PascalCase
        def __init__(self, n_components: int, **kwargs):  # type: ignore[no-untyped-def]
            self.n_components = int(n_components)
            if behavior.include_startprob:
                if behavior.startprob_values:
                    values = list(behavior.startprob_values)
                    if len(values) < self.n_components:
                        values.extend([0.0 for _ in range(self.n_components - len(values))])
                    self.startprob_ = values[: self.n_components]
                else:
                    self.startprob_ = [0.0 for _ in range(self.n_components)]
            if behavior.include_transmat:
                self.transmat_ = [
                    [0.0 for _ in range(self.n_components)] for _ in range(self.n_components)
                ]

        def fit(self, X, lengths=None):  # type: ignore[no-untyped-def]
            sequence = list(behavior.predicted_states or [])
            if not sequence:
                sequence = [0 for _ in range(len(X))]
            if len(sequence) < len(X):
                expanded: List[int] = []
                for idx in range(len(X)):
                    expanded.append(sequence[idx % len(sequence)])
                sequence = expanded
            counts: Dict[tuple[int, int], int] = {}
            totals: Dict[int, int] = {}
            for prev, nxt in zip(sequence, sequence[1:]):
                counts[(prev, nxt)] = counts.get((prev, nxt), 0) + 1
                totals[prev] = totals.get(prev, 0) + 1
            if behavior.include_transmat:
                for from_state in range(self.n_components):
                    denom = totals.get(from_state, 0) or 1
                    for to_state in range(self.n_components):
                        count = counts.get((from_state, to_state), 0)
                        self.transmat_[from_state][to_state] = float(count) / float(denom)
            return self

        def predict(self, X, lengths=None):  # type: ignore[no-untyped-def]
            sequence = list(behavior.predicted_states or [])
            if not sequence:
                sequence = [0 for _ in range(len(X))]
            if len(sequence) < len(X):
                expanded: List[int] = []
                for idx in range(len(X)):
                    expanded.append(sequence[idx % len(sequence)])
                sequence = expanded
            return sequence[: len(X)]

    class CategoricalHMM(GaussianHMM):  # noqa: N801 - external dependency uses PascalCase
        pass

    hmm_module = types.ModuleType("hmmlearn.hmm")
    hmm_module.GaussianHMM = GaussianHMM
    hmm_module.CategoricalHMM = CategoricalHMM

    root = types.ModuleType("hmmlearn")
    root.hmm = hmm_module

    sys.modules["hmmlearn"] = root
    sys.modules["hmmlearn.hmm"] = hmm_module

    context._fake_hmmlearn_installed = True
    context._fake_hmmlearn_original_modules = original_modules


def _install_hmmlearn_unavailable_module(context) -> None:
    already_installed = getattr(context, "_fake_hmmlearn_unavailable_installed", False)
    if already_installed:
        return

    original_modules: Dict[str, object] = {}
    for name in ["hmmlearn", "hmmlearn.hmm"]:
        if name in sys.modules:
            original_modules[name] = sys.modules[name]

    sys.modules["hmmlearn"] = types.ModuleType("hmmlearn")
    sys.modules.pop("hmmlearn.hmm", None)

    context._fake_hmmlearn_unavailable_installed = True
    context._fake_hmmlearn_unavailable_original_modules = original_modules


def _corpus_path(context, name: str) -> Path:
    return (context.workdir / name).resolve()


def _parse_json_output(standard_output: str) -> dict[str, object]:
    return json.loads(standard_output)


def _analysis_snapshot_dir(context, corpus_name: str = "corpus") -> Path:
    output = context.last_analysis_output
    snapshot_id = output["snapshot"]["snapshot_id"]
    corpus_path = _corpus_path(context, corpus_name)
    corpus = Corpus.open(corpus_path)
    return corpus.analysis_run_dir(analysis_id="markov", snapshot_id=snapshot_id)


def _load_markov_segments(context, corpus_name: str = "corpus") -> List[dict[str, object]]:
    path = _analysis_snapshot_dir(context, corpus_name) / "segments.jsonl"
    lines = [line for line in path.read_text(encoding="utf-8").splitlines() if line.strip()]
    return [json.loads(line) for line in lines]


def _load_markov_observations(context, corpus_name: str = "corpus") -> List[dict[str, object]]:
    path = _analysis_snapshot_dir(context, corpus_name) / "observations.jsonl"
    lines = [line for line in path.read_text(encoding="utf-8").splitlines() if line.strip()]
    return [json.loads(line) for line in lines]


def _latest_extraction_snapshot_manifest_path(context, corpus_name: str) -> Path:
    corpus = _corpus_path(context, corpus_name)
    extractor_id = context.last_extractor_id
    snapshot_id = context.last_extraction_snapshot_id
    assert isinstance(extractor_id, str) and extractor_id
    assert isinstance(snapshot_id, str) and snapshot_id
    return (
        corpus
        / "extracted"
        / extractor_id
        / snapshot_id
        / "manifest.json"
    )


def _latest_extraction_snapshot_root(context, corpus_name: str) -> Path:
    return _latest_extraction_snapshot_manifest_path(context, corpus_name).parent


def _snapshot_reference_from_context(context) -> str:
    extractor_id = context.last_extractor_id
    snapshot_id = context.last_extraction_snapshot_id
    assert isinstance(extractor_id, str) and extractor_id
    assert isinstance(snapshot_id, str) and snapshot_id
    return f"{extractor_id}:{snapshot_id}"


@given('a fake hmmlearn library is available with predicted states "{states}"')
def step_fake_hmmlearn_available(context, states: str) -> None:
    _install_fake_hmmlearn_module(context)
    behavior = _ensure_fake_hmmlearn_behavior(context)
    parsed: List[int] = []
    for token in (states or "").split(","):
        token = token.strip()
        if not token:
            continue
        parsed.append(int(token))
    behavior.predicted_states = parsed
    behavior.include_startprob = True
    behavior.startprob_values = None


@given("a fake hmmlearn library is available without transmat_ output")
def step_fake_hmmlearn_without_transmat(context) -> None:
    _install_fake_hmmlearn_module(context)
    behavior = _ensure_fake_hmmlearn_behavior(context)
    behavior.include_transmat = False


@given('a fake hmmlearn library is available with start probabilities "{values}"')
def step_fake_hmmlearn_with_startprob(context, values: str) -> None:
    _install_fake_hmmlearn_module(context)
    behavior = _ensure_fake_hmmlearn_behavior(context)
    parsed: List[float] = []
    for token in (values or "").split(","):
        token = token.strip()
        if not token:
            continue
        parsed.append(float(token))
    behavior.startprob_values = parsed
    behavior.include_startprob = True


@given("a fake hmmlearn library is available without start probabilities")
def step_fake_hmmlearn_without_startprob(context) -> None:
    _install_fake_hmmlearn_module(context)
    behavior = _ensure_fake_hmmlearn_behavior(context)
    behavior.include_startprob = False


@given("the hmmlearn dependency is unavailable")
def step_hmmlearn_dependency_unavailable(context) -> None:
    _install_hmmlearn_unavailable_module(context)


@when(
    'I snapshot a markov analysis in corpus "{corpus_name}" using configuration "{configuration_file}" '
    "and the latest extraction snapshot"
)
def step_run_markov_analysis_with_latest_extraction(
    context, corpus_name: str, configuration_file: str
) -> None:
    corpus = _corpus_path(context, corpus_name)
    workdir = getattr(context, "workdir", None)
    assert workdir is not None
    configuration_path = Path(workdir) / configuration_file
    snapshot_ref = _snapshot_reference_from_context(context)
    args = [
        "--corpus",
        str(corpus),
        "analyze",
        "markov",
        "--configuration",
        str(configuration_path),
        "--extraction-snapshot",
        snapshot_ref,
    ]
    result = run_biblicus(context, args, extra_env=getattr(context, "extra_env", None))
    context.last_result = result
    if result.returncode != 0:
        raise AssertionError(f"Markov analysis failed. stderr: {result.stderr}")
    context.last_analysis_output = _parse_json_output(result.stdout)


@when(
    'I attempt to snapshot a markov analysis in corpus "{corpus_name}" using configuration "{configuration_file}" '
    "and the latest extraction snapshot"
)
def step_attempt_run_markov_analysis_with_latest_extraction(
    context, corpus_name: str, configuration_file: str
) -> None:
    corpus = _corpus_path(context, corpus_name)
    workdir = getattr(context, "workdir", None)
    assert workdir is not None
    configuration_path = Path(workdir) / configuration_file
    snapshot_ref = _snapshot_reference_from_context(context)
    args = [
        "--corpus",
        str(corpus),
        "analyze",
        "markov",
        "--configuration",
        str(configuration_path),
        "--extraction-snapshot",
        snapshot_ref,
    ]
    result = run_biblicus(context, args, extra_env=getattr(context, "extra_env", None))
    context.last_result = result
    if result.returncode == 0:
        context.last_analysis_output = _parse_json_output(result.stdout)


@when(
    'I snapshot a markov analysis in corpus "{corpus_name}" using configuration "{configuration_file}"'
)
def step_run_markov_analysis_without_extraction_snapshot(
    context, corpus_name: str, configuration_file: str
) -> None:
    corpus = _corpus_path(context, corpus_name)
    workdir = getattr(context, "workdir", None)
    assert workdir is not None
    configuration_path = Path(workdir) / configuration_file
    args = [
        "--corpus",
        str(corpus),
        "analyze",
        "markov",
        "--configuration",
        str(configuration_path),
    ]
    result = run_biblicus(context, args, extra_env=getattr(context, "extra_env", None))
    context.last_result = result
    if result.returncode == 0:
        context.last_analysis_output = _parse_json_output(result.stdout)


@then("the markov analysis output includes {count:d} states")
def step_markov_output_includes_state_count(context, count: int) -> None:
    output = context.last_analysis_output
    states = output["report"]["states"]
    assert len(states) == count


@then("the markov analysis output references an extraction snapshot with artifacts")
def step_markov_output_references_extraction_snapshot(context) -> None:
    output = context.last_analysis_output
    snapshot = output["snapshot"]["input"]["extraction_snapshot"]
    extractor_id = snapshot["extractor_id"]
    snapshot_id = snapshot["snapshot_id"]
    corpus_path = _corpus_path(context, "corpus")
    manifest_path = (
        corpus_path / "extracted" / extractor_id / snapshot_id / "manifest.json"
    )
    assert manifest_path.is_file(), manifest_path


@then(
    "the markov analysis output includes a transition from state {from_state:d} to state {to_state:d}"
)
def step_markov_output_includes_transition(context, from_state: int, to_state: int) -> None:
    output = context.last_analysis_output
    transitions = output["report"]["transitions"]
    edges = {(edge["from_state"], edge["to_state"]) for edge in transitions}
    assert (from_state, to_state) in edges


@then("the markov analysis output includes {count:d} decoded item path")
def step_markov_output_includes_decoded_path_count(context, count: int) -> None:
    output = context.last_analysis_output
    paths = output["report"]["decoded_paths"]
    assert len(paths) == count


@then("the markov analysis output includes a topic modeling report")
def step_markov_output_includes_topic_modeling(context) -> None:
    output = context.last_analysis_output
    assert output["report"]["topic_modeling"] is not None


@then("the analysis snapshot includes a graphviz transitions file")
def step_markov_run_includes_graphviz_file(context) -> None:
    path = _analysis_snapshot_dir(context) / "transitions.dot"
    assert path.is_file()


@then("the markov analysis snapshot includes more than 1 segment")
def step_markov_run_includes_more_than_one_segment(context) -> None:
    segments = _load_markov_segments(context)
    assert len(segments) > 1


@then("the markov analysis snapshot includes a segment with text:")
def step_markov_run_includes_segment_text(context) -> None:
    expected = str(getattr(context, "text", "") or "").strip()
    segments = _load_markov_segments(context)
    assert any(str(segment.get("text", "")).strip() == expected for segment in segments)


@then("the markov analysis snapshot includes an observations file")
def step_markov_run_includes_observations_file(context) -> None:
    path = _analysis_snapshot_dir(context) / "observations.jsonl"
    assert path.is_file()


@then('the markov observations include topic labels "{labels}"')
def step_markov_observations_include_topic_labels(context, labels: str) -> None:
    expected = {token.strip() for token in labels.split(",") if token.strip()}
    observations = _load_markov_observations(context)
    found = {str(obs.get("topic_label", "")).strip() for obs in observations}
    for label in expected:
        assert label in found


@given("I append a non-extracted item to the latest extraction snapshot manifest")
@when("I append a non-extracted item to the latest extraction snapshot manifest")
def step_append_non_extracted_item_to_manifest(context) -> None:
    manifest_path = _latest_extraction_snapshot_manifest_path(context, "corpus")
    payload = json.loads(manifest_path.read_text(encoding="utf-8"))
    items = payload.get("items", [])
    assert isinstance(items, list)
    items.append(
        {
            "item_id": "item-non-extracted",
            "status": "failed",
            "final_text_relpath": None,
            "error_type": "failed",
            "error_message": "failed",
            "stage_results": [],
        }
    )
    payload["items"] = items
    manifest_path.write_text(json.dumps(payload, indent=2) + "\n", encoding="utf-8")


@given("I blank the extracted text for the 4th ingested item in the latest extraction snapshot")
@when("I blank the extracted text for the 4th ingested item in the latest extraction snapshot")
def step_blank_extracted_text_for_fourth_item(context) -> None:
    ingested_ids = getattr(context, "ingested_ids", None)
    assert isinstance(ingested_ids, list)
    assert len(ingested_ids) >= 4
    target_id = str(ingested_ids[3])

    manifest_path = _latest_extraction_snapshot_manifest_path(context, "corpus")
    payload = json.loads(manifest_path.read_text(encoding="utf-8"))
    items = payload.get("items", [])
    assert isinstance(items, list)
    target_item = None
    for item in items:
        if isinstance(item, dict) and item.get("item_id") == target_id:
            target_item = item
            break
    assert target_item is not None
    relpath = target_item.get("final_text_relpath")
    assert isinstance(relpath, str) and relpath
    text_path = _latest_extraction_snapshot_root(context, "corpus") / relpath
    text_path.write_text("", encoding="utf-8")


@then("the markov analysis text collection report includes:")
def step_markov_text_collection_report_includes(context) -> None:
    output = context.last_analysis_output
    report = output["report"]["text_collection"]
    for row in context.table:
        field = str(row["field"])
        expected = int(row["value"])
        assert int(report[field]) == expected


@then('the markov analysis text collection report warnings include "{snippet}"')
def step_markov_text_collection_report_warnings_include(context, snippet: str) -> None:
    output = context.last_analysis_output
    report = output["report"]["text_collection"]
    warnings = report.get("warnings", [])
    assert isinstance(warnings, list)
    combined = "\n".join(str(value) for value in warnings)
    assert snippet in combined


@then("every Markov state report has no non-boundary exemplars")
def step_every_markov_state_report_has_no_non_boundary_exemplars(context) -> None:
    output = context.last_analysis_output
    states = output["report"]["states"]
    for state in states:
        assert isinstance(state, dict)
        exemplars = state.get("exemplars", [])
        assert isinstance(exemplars, list)
        non_boundary = [
            exemplar
            for exemplar in exemplars
            if str(exemplar).strip().upper() not in {"START", "END"}
        ]
        assert non_boundary == []


@then("the graphviz transitions file contains no edges")
def step_graphviz_transitions_file_contains_no_edges(context) -> None:
    path = _analysis_snapshot_dir(context) / "transitions.dot"
    contents = path.read_text(encoding="utf-8")
    assert "->" not in contents



"""
Retriever registry for Biblicus retrieval engines.
"""

from __future__ import annotations

from typing import Dict, Type

from .base import Retriever
from .embedding_index_file import EmbeddingIndexFileRetriever
from .embedding_index_inmemory import EmbeddingIndexInMemoryRetriever
from .hybrid import HybridRetriever
from .scan import ScanRetriever
from .sqlite_full_text_search import SqliteFullTextSearchRetriever
from .tf_vector import TfVectorRetriever


def available_retrievers() -> Dict[str, Type[Retriever]]:
    """
    Return the registered retrievers.

    :return: Mapping of retriever identifiers to retriever classes.
    :rtype: dict[str, Type[Retriever]]
    """
    return {
        EmbeddingIndexFileRetriever.retriever_id: EmbeddingIndexFileRetriever,
        EmbeddingIndexInMemoryRetriever.retriever_id: EmbeddingIndexInMemoryRetriever,
        HybridRetriever.retriever_id: HybridRetriever,
        ScanRetriever.retriever_id: ScanRetriever,
        SqliteFullTextSearchRetriever.retriever_id: SqliteFullTextSearchRetriever,
        TfVectorRetriever.retriever_id: TfVectorRetriever,
    }


def get_retriever(retriever_id: str) -> Retriever:
    """
    Instantiate a retriever by identifier.

    :param retriever_id: Retriever identifier.
    :type retriever_id: str
    :return: Retriever instance.
    :rtype: Retriever
    :raises KeyError: If the retriever identifier is unknown.
    """
    registry = available_retrievers()
    retriever_class = registry.get(retriever_id)
    if retriever_class is None:
        known = ", ".join(sorted(registry))
        raise KeyError(f"Unknown retriever '{retriever_id}'. Known retrievers: {known}")
    return retriever_class()



"""
Remote collection mirroring for Biblicus.
"""

from __future__ import annotations

import json
import shutil
from pathlib import Path
from typing import Iterable, List

from .corpus import Corpus
from .models import (
    CollectionMembership,
    CorpusConfig,
    RemoteCollectionPullResult,
    RemoteCorpusCollectionConfig,
    RemoteCorpusCollectionDiscovery,
    RemoteCorpusSourceConfig,
)
from .remote_sources import AzureBlobRemoteSource, S3RemoteSource
from .time import utc_now_iso
from .user_config import resolve_source_profile

COLLECTION_CONFIG_NAME = "config.json"
COLLECTION_DIR_NAME = "collections"
ARCHIVE_DIR_NAME = ".archived"


def collection_config_path(collection_root: Path) -> Path:
    """
    Return the path to the collection configuration file.

    :param collection_root: Collection root directory.
    :type collection_root: Path
    :return: Configuration file path.
    :rtype: Path
    """
    return collection_root / "metadata" / COLLECTION_CONFIG_NAME


def load_collection_config(collection_root: Path) -> RemoteCorpusCollectionConfig:
    """
    Load the collection configuration from disk.

    :param collection_root: Collection root directory.
    :type collection_root: Path
    :return: Parsed collection configuration.
    :rtype: RemoteCorpusCollectionConfig
    :raises FileNotFoundError: If the config file is missing.
    :raises ValueError: If the config file is invalid.
    """
    config_path = collection_config_path(collection_root)
    if not config_path.is_file():
        raise FileNotFoundError(f"Missing collection config: {config_path}")
    raw = json.loads(config_path.read_text(encoding="utf-8"))
    return RemoteCorpusCollectionConfig.model_validate(raw)


def init_collection(collection_root: Path, config: RemoteCorpusCollectionConfig) -> None:
    """
    Initialize a collection directory on disk.

    :param collection_root: Collection root directory.
    :type collection_root: Path
    :param config: Collection configuration to persist.
    :type config: RemoteCorpusCollectionConfig
    :return: None.
    :rtype: None
    """
    metadata_dir = collection_root / "metadata"
    metadata_dir.mkdir(parents=True, exist_ok=True)
    config_path = metadata_dir / COLLECTION_CONFIG_NAME
    config_path.write_text(config.model_dump_json(indent=2) + "\n", encoding="utf-8")


def pull_collection(collection_root: Path) -> RemoteCollectionPullResult:
    """
    Mirror a remote collection into local corpora.

    :param collection_root: Collection root directory.
    :type collection_root: Path
    :return: Collection pull summary.
    :rtype: RemoteCollectionPullResult
    """
    collection_root = collection_root.resolve()
    config = load_collection_config(collection_root)
    profile = resolve_source_profile(config.source.profile)
    if config.source.kind != profile.kind:
        raise ValueError(
            "Collection source kind does not match source profile kind: "
            f"{config.source.kind} vs {profile.kind}"
        )

    if config.source.kind == "s3":
        source = S3RemoteSource(config.source, profile)
    elif config.source.kind == "azure-blob":
        source = AzureBlobRemoteSource(config.source, profile)
    else:
        raise ValueError(f"Unsupported remote source kind: {config.source.kind}")

    result = RemoteCollectionPullResult()
    discovery = config.discovery
    corpus_root = _resolve_corpus_root(collection_root, config)

    if discovery.mode == "partition":
        corpus = _ensure_partition_corpus(
            corpus_root,
            collection_name=config.collection_name,
            source_config=_build_partition_source_config(config.source),
        )
        tag_resolver = _partition_tag_resolver()
        mirror = corpus.pull_source(tag_resolver=tag_resolver)
        result.discovered = len(_discover_subfolders(source, config.source, discovery))
        result.mirrored = 1 if mirror.listed > 0 else 0
        return result

    subfolders = _discover_subfolders(source, config.source, discovery)
    result.discovered = len(subfolders)

    created = 0
    mirrored = 0
    for subfolder in subfolders:
        corpus_path = corpus_root / subfolder
        created += int(
            _ensure_collection_corpus(
                corpus_path,
                collection_name=config.collection_name,
                corpus_name=subfolder,
                source_config=_build_subfolder_source_config(config.source, subfolder),
                auto_create=config.auto_create,
            )
        )
        corpus = Corpus.open(corpus_path)
        corpus.pull_source()
        mirrored += 1

    result.created = created
    result.mirrored = mirrored
    result.archived = _archive_missing_corpora(
        corpus_root, subfolders, deletion_policy=config.deletion_policy
    )
    return result


def _resolve_corpus_root(collection_root: Path, config: RemoteCorpusCollectionConfig) -> Path:
    base = collection_root.parent
    if base.name == COLLECTION_DIR_NAME:
        base = base.parent
    corpus_root = Path(config.corpus_root)
    if corpus_root.is_absolute():
        return corpus_root
    return (base / corpus_root).resolve()


def _build_subfolder_source_config(
    source_config: RemoteCorpusSourceConfig, subfolder: str
) -> RemoteCorpusSourceConfig:
    prefix = _join_prefix(source_config.prefix, subfolder)
    return RemoteCorpusSourceConfig(
        kind=source_config.kind,
        profile=source_config.profile,
        name=subfolder,
        bucket=source_config.bucket,
        container=source_config.container,
        prefix=prefix,
    )


def _build_partition_source_config(
    source_config: RemoteCorpusSourceConfig,
) -> RemoteCorpusSourceConfig:
    return RemoteCorpusSourceConfig(
        kind=source_config.kind,
        profile=source_config.profile,
        name=source_config.name,
        bucket=source_config.bucket,
        container=source_config.container,
        prefix=source_config.prefix,
    )


def _ensure_collection_corpus(
    corpus_root: Path,
    *,
    collection_name: str,
    corpus_name: str,
    source_config: RemoteCorpusSourceConfig,
    auto_create: bool,
) -> bool:
    config_path = corpus_root / "metadata" / "config.json"
    created = False
    if not config_path.exists():
        if not auto_create:
            raise ValueError(f"Corpus missing for collection: {corpus_root}")
        Corpus.init(corpus_root, force=True)
        created = True
    else:
        _ = Corpus.open(corpus_root)

    base_config = json.loads(config_path.read_text(encoding="utf-8"))
    config = CorpusConfig.model_validate(base_config)
    updated = config.model_copy(
        update={
            "source": source_config,
            "collection": CollectionMembership(
                collection_name=collection_name,
                corpus_name=corpus_name,
            ),
        }
    )
    config_path.write_text(updated.model_dump_json(indent=2) + "\n", encoding="utf-8")
    return created


def _ensure_partition_corpus(
    corpus_root: Path,
    *,
    collection_name: str,
    source_config: RemoteCorpusSourceConfig,
) -> Corpus:
    config_path = corpus_root / "metadata" / "config.json"
    if not config_path.exists():
        Corpus.init(corpus_root, force=True)
    base_config = json.loads(config_path.read_text(encoding="utf-8"))
    config = CorpusConfig.model_validate(base_config)
    updated = config.model_copy(
        update={
            "source": source_config,
            "collection": CollectionMembership(
                collection_name=collection_name,
                corpus_name=collection_name,
            ),
        }
    )
    config_path.write_text(updated.model_dump_json(indent=2) + "\n", encoding="utf-8")
    return Corpus.open(corpus_root)


def _discover_subfolders(
    source: object,
    source_config: RemoteCorpusSourceConfig,
    discovery: RemoteCorpusCollectionDiscovery,
) -> List[str]:
    items = source.list_items()
    subfolders = set()
    for item in items:
        relative = _relative_key(item.key, source_config.prefix)
        if not relative:
            continue
        parts = [segment for segment in relative.split("/") if segment]
        if len(parts) < 2:
            continue
        subfolders.add(parts[0])
    return sorted(subfolders)


def _partition_tag_resolver():
    def _resolver(relative_key: str) -> List[str]:
        if not relative_key:
            return []
        parts = [segment for segment in relative_key.split("/") if segment]
        if len(parts) < 2:
            return []
        return [f"table:{parts[0]}"]

    return _resolver


def _archive_missing_corpora(
    corpus_root: Path, discovered: Iterable[str], *, deletion_policy: str
) -> int:
    discovered_set = set(discovered)
    if not corpus_root.exists():
        return 0
    archived = 0
    for child in corpus_root.iterdir():
        if not child.is_dir():
            continue
        if child.name == ARCHIVE_DIR_NAME:
            continue
        if child.name in discovered_set:
            continue
        if deletion_policy == "delete":
            shutil.rmtree(child)
            archived += 1
        else:
            archive_root = corpus_root / ARCHIVE_DIR_NAME
            archive_root.mkdir(parents=True, exist_ok=True)
            destination = archive_root / f"{child.name}-{_archive_suffix()}"
            child.rename(destination)
            archived += 1
    return archived


def _archive_suffix() -> str:
    return utc_now_iso().replace(":", "").replace("-", "")


def _relative_key(key: str, prefix: str) -> str:
    relative = key
    if prefix and relative.startswith(prefix):
        relative = relative[len(prefix) :]
    return relative.lstrip("/")


def _join_prefix(prefix: str, subfolder: str) -> str:
    if prefix:
        return f"{prefix.rstrip('/')}/{subfolder.strip('/')}/"
    return f"{subfolder.strip('/')}/"



from __future__ import annotations

import json

from behave import then, when
from pydantic import ValidationError

from biblicus.analysis.models import (
    MarkovAnalysisConfiguration,
    MarkovAnalysisModelFamily,
    MarkovAnalysisObservationsEncoder,
    MarkovAnalysisSegmentationMethod,
)


@when(
    'I attempt to validate a Markov configuration with segmentation method "{method}" and no llm config'
)
def step_attempt_validate_markov_configuration_missing_llm(context, method: str) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {"method": method},
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    'I attempt to validate a Markov configuration with segmentation method "{method}" and no span markup config'
)
def step_attempt_validate_markov_configuration_missing_span_markup(context, method: str) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {"method": method},
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    'I attempt to validate a Markov configuration with span markup system prompt missing "{token}"'
)
def step_attempt_validate_markov_configuration_span_markup_system_prompt_missing(
    context, token: str
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {
                    "method": "span_markup",
                    "span_markup": {
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "prompt_template": "Return the segments.",
                        "system_prompt": f"Missing token {token.strip('{}')}",
                    },
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    'I attempt to validate a Markov configuration with span markup prompt template containing "{token}"'
)
def step_attempt_validate_markov_configuration_span_markup_prompt_contains(
    context, token: str
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {
                    "method": "span_markup",
                    "span_markup": {
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "prompt_template": f"Return {token}.",
                        "system_prompt": "Current text:\n---\n{text}\n---\n",
                    },
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with span markup prepend label enabled and no label attribute"
)
def step_attempt_validate_markov_configuration_span_markup_missing_label_attribute(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {
                    "method": "span_markup",
                    "span_markup": {
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "prompt_template": "Return the segments.",
                        "system_prompt": "Current text:\n---\n{text}\n---\n",
                        "prepend_label": True,
                    },
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with span markup end label value but no verifier"
)
def step_attempt_validate_markov_configuration_span_markup_end_label_missing_verifier(
    context,
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {
                    "method": "span_markup",
                    "span_markup": {
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "prompt_template": "Return the segments.",
                        "system_prompt": "Current text:\n---\n{text}\n---\n",
                        "end_label_value": "End",
                    },
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with span markup end reject label value but no verifier"
)
def step_attempt_validate_markov_configuration_span_markup_end_reject_missing_verifier(
    context,
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {
                    "method": "span_markup",
                    "span_markup": {
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "prompt_template": "Return the segments.",
                        "system_prompt": "Current text:\n---\n{text}\n---\n",
                        "end_reject_label_value": "Reject",
                    },
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    'I attempt to validate a Markov configuration with end label verifier system prompt missing "{text_placeholder}"'
)
def step_attempt_validate_markov_configuration_end_verifier_system_prompt_missing(
    context, text_placeholder: str
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {
                    "method": "span_markup",
                    "span_markup": {
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "prompt_template": "Return the segments.",
                        "system_prompt": "Current text:\n---\n{text}\n---\n",
                        "end_label_value": "End",
                        "end_label_verifier": {
                            "client": {
                                "provider": "openai",
                                "model": "gpt-4o-mini",
                                "api_key": "test-key",
                            },
                            "system_prompt": "Missing placeholder.",
                            "prompt_template": "Verify end label.",
                        },
                    },
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    'I attempt to validate a Markov configuration with end label verifier prompt template containing "{text_placeholder}"'
)
def step_attempt_validate_markov_configuration_end_verifier_prompt_contains(
    context, text_placeholder: str
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {
                    "method": "span_markup",
                    "span_markup": {
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "prompt_template": "Return the segments.",
                        "system_prompt": "Current text:\n---\n{text}\n---\n",
                        "end_label_value": "End",
                        "end_label_verifier": {
                            "client": {
                                "provider": "openai",
                                "model": "gpt-4o-mini",
                                "api_key": "test-key",
                            },
                            "system_prompt": "Verify end label:\n{text}",
                            "prompt_template": f"Verify {text_placeholder}",
                        },
                    },
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with state naming enabled and missing requirements"
)
def step_attempt_validate_markov_configuration_state_naming_missing(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "report": {"state_naming": {"enabled": True}},
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with state naming enabled and missing system prompt"
)
def step_attempt_validate_markov_configuration_state_naming_missing_system_prompt(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "report": {
                    "state_naming": {
                        "enabled": True,
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "prompt_template": "Name the state.",
                    }
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with state naming enabled and missing prompt template"
)
def step_attempt_validate_markov_configuration_state_naming_missing_prompt_template(
    context,
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "report": {
                    "state_naming": {
                        "enabled": True,
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "system_prompt": "Context:\n{context_pack}",
                    }
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when("I validate a Markov configuration with state naming disabled")
def step_validate_markov_configuration_state_naming_disabled(context) -> None:
    config = MarkovAnalysisConfiguration.model_validate(
        {
            "schema_version": 1,
            "report": {"state_naming": {"enabled": False}},
            "model": {"family": "gaussian", "n_states": 2},
            "observations": {"encoder": "tfidf"},
        }
    )
    context.last_markov_configuration = config


@then("the Markov state naming is disabled")
def step_markov_state_naming_is_disabled(context) -> None:
    config = getattr(context, "last_markov_configuration", None)
    assert config is not None
    state_naming = config.report.state_naming
    assert state_naming is not None
    assert state_naming.enabled is False


@when(
    'I attempt to validate a Markov configuration with state naming system prompt missing "{context_pack}"'
)
def step_attempt_validate_markov_configuration_state_naming_missing_context_pack(
    context, context_pack: str
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "report": {
                    "state_naming": {
                        "enabled": True,
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "system_prompt": "Missing context pack placeholder.",
                        "prompt_template": "Name the state.",
                    }
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    'I attempt to validate a Markov configuration with state naming prompt template containing "{context_pack}"'
)
def step_attempt_validate_markov_configuration_state_naming_prompt_contains_context_pack(
    context, context_pack: str
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "report": {
                    "state_naming": {
                        "enabled": True,
                        "client": {
                            "provider": "openai",
                            "model": "gpt-4o-mini",
                            "api_key": "test-key",
                        },
                        "system_prompt": f"Context: {context_pack}",
                        "prompt_template": f"Name {context_pack}",
                    }
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with topic modeling enabled and no configuration"
)
def step_attempt_validate_markov_configuration_topic_modeling_enabled_without_configuration(
    context,
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "topic_modeling": {"enabled": True},
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with topic modeling enabled and non-single LLM extraction"
)
def step_attempt_validate_markov_configuration_topic_modeling_non_single_llm_extraction(
    context,
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "topic_modeling": {
                    "enabled": True,
                    "configuration": {
                        "schema_version": 1,
                        "llm_extraction": {
                            "enabled": True,
                            "method": "itemize",
                            "client": {
                                "provider": "openai",
                                "model": "gpt-4o-mini",
                                "api_key": "test-key",
                            },
                            "prompt_template": "{text}",
                        },
                    },
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when("I validate a Markov configuration with span markup prompts")
def step_validate_markov_configuration_span_markup_prompts(context) -> None:
    configuration = MarkovAnalysisConfiguration.model_validate(
        {
            "schema_version": 1,
            "segmentation": {
                "method": "span_markup",
                "span_markup": {
                    "client": {
                        "provider": "openai",
                        "model": "gpt-4o-mini",
                        "api_key": "test-key",
                    },
                    "prompt_template": "Return the segments.",
                    "system_prompt": "Current text:\n---\n{text}\n---\n",
                },
            },
            "model": {"family": "gaussian", "n_states": 2},
            "observations": {"encoder": "tfidf"},
        }
    )
    context.markov_configuration = configuration
    context.last_markov_configuration = configuration


@when(
    "I attempt to validate a Markov configuration with llm observations enabled and missing requirements"
)
def step_attempt_validate_markov_configuration_llm_observations_missing(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "llm_observations": {"enabled": True},
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with llm observations enabled and empty prompt template"
)
def step_attempt_validate_markov_configuration_llm_observations_empty_prompt(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "llm_observations": {
                    "enabled": True,
                    "client": {"provider": "openai", "model": "gpt-4o-mini", "api_key": "test-key"},
                    "prompt_template": "",
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when(
    "I attempt to validate a Markov configuration with embeddings enabled and invalid text source"
)
def step_attempt_validate_markov_configuration_embeddings_invalid_source(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "embeddings": {
                    "enabled": True,
                    "text_source": "unknown",
                    "client": {
                        "provider": "openai",
                        "model": "text-embedding-3-small",
                        "api_key": "test-key",
                    },
                },
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when("I attempt to validate a Markov configuration with embeddings enabled and missing client")
def step_attempt_validate_markov_configuration_embeddings_missing_client(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "embeddings": {"enabled": True, "text_source": "segment_text"},
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when('I attempt to validate a Markov configuration with observations text_source "{source}"')
def step_attempt_validate_markov_configuration_observations_text_source(
    context, source: str
) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "observations": {"encoder": "tfidf", "text_source": source},
                "model": {"family": "gaussian", "n_states": 2},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when("I attempt to validate a Markov configuration with schema version {schema_version:d}")
def step_attempt_validate_markov_configuration_schema_version(context, schema_version: int) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": schema_version,
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when("I validate a Markov configuration with enum segmentation method")
def step_validate_markov_configuration_enum_segmentation_method(context) -> None:
    configuration = MarkovAnalysisConfiguration.model_validate(
        {
            "schema_version": 1,
            "segmentation": {"method": MarkovAnalysisSegmentationMethod.SENTENCE},
            "model": {"family": "gaussian", "n_states": 2},
            "observations": {"encoder": "tfidf"},
        }
    )
    context.last_markov_configuration = configuration


@then('the Markov segmentation method equals "{expected}"')
def step_markov_segmentation_method_equals(context, expected: str) -> None:
    configuration = getattr(context, "last_markov_configuration", None)
    assert configuration is not None
    assert configuration.segmentation.method.value == expected


@when("I attempt to validate a Markov configuration with invalid segmentation method type")
def step_attempt_validate_markov_configuration_invalid_segmentation_method_type(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "segmentation": {"method": 123},
                "model": {"family": "gaussian", "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when("I validate a Markov configuration with enum observations encoder")
def step_validate_markov_configuration_enum_observations_encoder(context) -> None:
    configuration = MarkovAnalysisConfiguration.model_validate(
        {
            "schema_version": 1,
            "observations": {"encoder": MarkovAnalysisObservationsEncoder.TFIDF},
            "model": {"family": "gaussian", "n_states": 2},
        }
    )
    context.last_markov_configuration = configuration


@then('the Markov observations encoder equals "{expected}"')
def step_markov_observations_encoder_equals(context, expected: str) -> None:
    configuration = getattr(context, "last_markov_configuration", None)
    assert configuration is not None
    assert configuration.observations.encoder.value == expected


@when("I attempt to validate a Markov configuration with invalid observations encoder type")
def step_attempt_validate_markov_configuration_invalid_observations_encoder_type(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "observations": {"encoder": 123},
                "model": {"family": "gaussian", "n_states": 2},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when("I validate a Markov configuration with enum model family")
def step_validate_markov_configuration_enum_model_family(context) -> None:
    configuration = MarkovAnalysisConfiguration.model_validate(
        {
            "schema_version": 1,
            "model": {"family": MarkovAnalysisModelFamily.GAUSSIAN, "n_states": 2},
            "observations": {"encoder": "tfidf"},
        }
    )
    context.last_markov_configuration = configuration


@then('the Markov model family equals "{expected}"')
def step_markov_model_family_equals(context, expected: str) -> None:
    configuration = getattr(context, "last_markov_configuration", None)
    assert configuration is not None
    assert configuration.model.family.value == expected


@when("I attempt to validate a Markov configuration with invalid model family type")
def step_attempt_validate_markov_configuration_invalid_model_family_type(context) -> None:
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "model": {"family": 123, "n_states": 2},
                "observations": {"encoder": "tfidf"},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc


@when("I attempt to validate a Markov configuration with tfidf ngram_range {raw_json}")
def step_attempt_validate_markov_configuration_tfidf_ngram_range(context, raw_json: str) -> None:
    value = json.loads(raw_json)
    try:
        MarkovAnalysisConfiguration.model_validate(
            {
                "schema_version": 1,
                "observations": {"encoder": "tfidf", "tfidf": {"ngram_range": value}},
                "model": {"family": "gaussian", "n_states": 2},
            }
        )
        context.validation_error = None
    except ValidationError as exc:
        context.validation_error = exc



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from pydantic import ValidationError

from biblicus.context_engine import (
    CompactorDeclaration,
    ContextDeclaration,
    CorpusDeclaration,
    RetrieverDeclaration,
)


@dataclass
class ContextRegistry:
    """
    Registry container for Context engine test declarations.

    :ivar contexts: Registered Context declarations.
    :vartype contexts: dict[str, ContextDeclaration]
    :ivar retrievers: Registered Retriever declarations.
    :vartype retrievers: dict[str, RetrieverDeclaration]
    :ivar corpora: Registered Corpus declarations.
    :vartype corpora: dict[str, CorpusDeclaration]
    :ivar compactors: Registered Compactor declarations.
    :vartype compactors: dict[str, CompactorDeclaration]
    """

    contexts: dict[str, ContextDeclaration] = field(default_factory=dict)
    retrievers: dict[str, RetrieverDeclaration] = field(default_factory=dict)
    corpora: dict[str, CorpusDeclaration] = field(default_factory=dict)
    compactors: dict[str, CompactorDeclaration] = field(default_factory=dict)


class RegistryBuilder:
    """
    Simple builder for Context engine registry tests.
    """

    def __init__(self) -> None:
        """
        Initialize an empty registry.
        """
        self.registry = ContextRegistry()

    def register_context(self, name: str, config: dict[str, Any]) -> None:
        """
        Register a Context declaration for tests.

        :param name: Context name.
        :type name: str
        :param config: Context configuration payload.
        :type config: dict[str, Any]
        :return: None.
        :rtype: None
        """
        context_config = dict(config)
        context_config["name"] = name
        try:
            self.registry.contexts[name] = ContextDeclaration(**context_config)
        except ValidationError as exc:
            raise AssertionError(f"Invalid context '{name}': {exc}") from exc

    def register_corpus(self, name: str, config: dict[str, Any]) -> None:
        """
        Register a Corpus declaration for tests.

        :param name: Corpus name.
        :type name: str
        :param config: Corpus configuration payload.
        :type config: dict[str, Any]
        :return: None.
        :rtype: None
        """
        corpus_config = dict(config)
        try:
            self.registry.corpora[name] = CorpusDeclaration(name=name, config=corpus_config)
        except ValidationError as exc:
            raise AssertionError(f"Invalid corpus '{name}': {exc}") from exc

    def register_retriever(self, name: str, config: dict[str, Any]) -> None:
        """
        Register a Retriever declaration for tests.

        :param name: Retriever name.
        :type name: str
        :param config: Retriever configuration payload.
        :type config: dict[str, Any]
        :return: None.
        :rtype: None
        """
        retriever_config = dict(config)
        corpus_name = retriever_config.pop("corpus", None)
        try:
            self.registry.retrievers[name] = RetrieverDeclaration(
                name=name,
                corpus=corpus_name,
                config=retriever_config,
            )
        except ValidationError as exc:
            raise AssertionError(f"Invalid retriever '{name}': {exc}") from exc

    def register_compactor(self, name: str, config: dict[str, Any]) -> None:
        """
        Register a Compactor declaration for tests.

        :param name: Compactor name.
        :type name: str
        :param config: Compactor configuration payload.
        :type config: dict[str, Any]
        :return: None.
        :rtype: None
        """
        compactor_config = dict(config)
        try:
            self.registry.compactors[name] = CompactorDeclaration(
                name=name,
                config=compactor_config,
            )
        except ValidationError as exc:
            raise AssertionError(f"Invalid compactor '{name}': {exc}") from exc



from __future__ import annotations

import os
import sys
import types
from dataclasses import dataclass
from typing import Any, Dict, Optional
from urllib.parse import unquote

from behave import given, then


@dataclass
class _FakeAzureRecognitionBehavior:
    transcript: str
    reason: str = "RecognizedSpeech"
    confidence: Optional[float] = None
    cancellation_reason: Optional[str] = None
    error_details: Optional[str] = None


def _ensure_fake_azure_recognition_behaviors(
    context,
) -> Dict[str, _FakeAzureRecognitionBehavior]:
    behaviors = getattr(context, "fake_azure_recognitions", None)
    if behaviors is None:
        behaviors = {}
        context.fake_azure_recognitions = behaviors
    return behaviors


def _install_fake_azure_speech_module(context) -> None:
    already_installed = getattr(context, "_fake_azure_speech_installed", False)
    if already_installed:
        # Reset module state even if already installed
        speechsdk_module = sys.modules.get("azure.cognitiveservices.speech")
        if speechsdk_module is not None:
            speechsdk_module.last_api_key = None
            speechsdk_module.last_region = None
            speechsdk_module.last_endpoint = None
            speechsdk_module.last_audio_filename = None
            speechsdk_module.last_speech_config = None
            speechsdk_module.last_audio_config = None
            speechsdk_module.last_profanity_option = None
            speechsdk_module.last_dictation_enabled = False
        return

    original_modules: Dict[str, object] = {}
    if "azure" in sys.modules:
        original_modules["azure"] = sys.modules["azure"]
    if "azure.cognitiveservices" in sys.modules:
        original_modules["azure.cognitiveservices"] = sys.modules["azure.cognitiveservices"]
    if "azure.cognitiveservices.speech" in sys.modules:
        original_modules["azure.cognitiveservices.speech"] = sys.modules["azure.cognitiveservices.speech"]

    behaviors = _ensure_fake_azure_recognition_behaviors(context)

    # Create nested module structure
    azure_module = types.ModuleType("azure")
    cognitiveservices_module = types.ModuleType("azure.cognitiveservices")
    speechsdk_module = types.ModuleType("azure.cognitiveservices.speech")

    # Link modules
    azure_module.cognitiveservices = cognitiveservices_module
    cognitiveservices_module.speech = speechsdk_module

    # Create enums for result reasons
    class ResultReason:
        RecognizedSpeech = "RecognizedSpeech"
        NoMatch = "NoMatch"
        Canceled = "Canceled"

    speechsdk_module.ResultReason = ResultReason

    class ProfanityOption:
        Masked = "Masked"
        Removed = "Removed"
        Raw = "Raw"

    speechsdk_module.ProfanityOption = ProfanityOption

    class _CancellationDetails:
        def __init__(self, reason: str, error_details: Optional[str] = None) -> None:
            self.reason = reason
            self.error_details = error_details

    class _RecognitionResult:
        def __init__(
            self,
            text: str,
            reason: str,
            confidence: Optional[float] = None,
            cancellation_reason: Optional[str] = None,
            error_details: Optional[str] = None
        ) -> None:
            self.text = text
            self.reason = reason
            self.confidence = confidence
            if cancellation_reason:
                self.cancellation_details = _CancellationDetails(cancellation_reason, error_details)

    class _AudioConfig:
        def __init__(self, filename: str) -> None:
            speechsdk_module.last_audio_filename = filename  # type: ignore[attr-defined]

    speechsdk_module.audio = types.ModuleType("audio")
    speechsdk_module.audio.AudioConfig = _AudioConfig

    class _SpeechConfig:
        def __init__(self, subscription: str, region: Optional[str] = None, endpoint: Optional[str] = None) -> None:
            speechsdk_module.last_api_key = subscription  # type: ignore[attr-defined]
            speechsdk_module.last_region = region  # type: ignore[attr-defined]
            speechsdk_module.last_endpoint = endpoint  # type: ignore[attr-defined]
            self.speech_recognition_language = None
            self._profanity = None
            self._dictation_enabled = False

        def set_profanity(self, option: str) -> None:
            self._profanity = option
            speechsdk_module.last_profanity_option = option  # type: ignore[attr-defined]

        def enable_dictation(self) -> None:
            self._dictation_enabled = True
            speechsdk_module.last_dictation_enabled = True  # type: ignore[attr-defined]

    speechsdk_module.SpeechConfig = _SpeechConfig

    class _SpeechRecognizer:
        def __init__(self, speech_config: _SpeechConfig, audio_config: _AudioConfig) -> None:
            speechsdk_module.last_speech_config = speech_config  # type: ignore[attr-defined]
            speechsdk_module.last_audio_config = audio_config  # type: ignore[attr-defined]

        def recognize_once(self) -> _RecognitionResult:
            filename = getattr(speechsdk_module, "last_audio_filename", "unknown")
            base_filename = filename.rsplit("/", 1)[-1]
            # Strip the UUID prefix (format: uuid--originalfile)
            base_name = base_filename.split("--", 1)[-1] if "--" in base_filename else base_filename
            # URL decode if needed
            base_name = unquote(base_name)
            # Extract just the basename (in case it's a full file:// URI)
            base_name = base_name.rsplit("/", 1)[-1]

            behavior = behaviors.get(base_name)
            if behavior is None:
                behavior = _FakeAzureRecognitionBehavior(transcript="")

            return _RecognitionResult(
                text=behavior.transcript,
                reason=behavior.reason,
                confidence=behavior.confidence,
                cancellation_reason=behavior.cancellation_reason,
                error_details=behavior.error_details
            )

    speechsdk_module.SpeechRecognizer = _SpeechRecognizer

    # Initialize/reset module attributes (including any leftover state from previous tests)
    speechsdk_module.last_api_key = None
    speechsdk_module.last_region = None
    speechsdk_module.last_endpoint = None
    speechsdk_module.last_audio_filename = None
    speechsdk_module.last_speech_config = None
    speechsdk_module.last_audio_config = None
    speechsdk_module.last_profanity_option = None
    speechsdk_module.last_dictation_enabled = False

    sys.modules["azure"] = azure_module
    sys.modules["azure.cognitiveservices"] = cognitiveservices_module
    sys.modules["azure.cognitiveservices.speech"] = speechsdk_module

    context._fake_azure_speech_installed = True
    context._fake_azure_speech_original_modules = original_modules


def _install_azure_speech_unavailable_module(context) -> None:
    already_installed = getattr(context, "_fake_azure_speech_unavailable_installed", False)
    if already_installed:
        return

    original_modules: Dict[str, object] = {}
    module_names = ["azure", "azure.cognitiveservices", "azure.cognitiveservices.speech"]
    for name in module_names:
        if name in sys.modules:
            original_modules[name] = sys.modules[name]

    # Create minimal module structure that will fail import check
    azure_module = types.ModuleType("azure")
    sys.modules["azure"] = azure_module

    context._fake_azure_speech_unavailable_installed = True
    context._fake_azure_speech_unavailable_original_modules = original_modules


@given("a fake Azure Speech library is available")
def step_fake_azure_speech_available(context) -> None:
    _install_fake_azure_speech_module(context)


@given(
    'a fake Azure Speech library is available that returns transcript "{transcript}" for filename "{filename}"'
)
def step_fake_azure_speech_returns_transcript(context, transcript: str, filename: str) -> None:
    _install_fake_azure_speech_module(context)
    behaviors = _ensure_fake_azure_recognition_behaviors(context)
    behaviors[filename] = _FakeAzureRecognitionBehavior(transcript=transcript)


@given(
    'a fake Azure Speech library is available that returns no speech for filename "{filename}"'
)
def step_fake_azure_speech_returns_no_speech(context, filename: str) -> None:
    _install_fake_azure_speech_module(context)
    behaviors = _ensure_fake_azure_recognition_behaviors(context)
    behaviors[filename] = _FakeAzureRecognitionBehavior(transcript="", reason="NoMatch")


@given(
    'a fake Azure Speech library is available that returns cancelled recognition for filename "{filename}" with reason "{reason}"'
)
def step_fake_azure_speech_returns_cancelled(context, filename: str, reason: str) -> None:
    _install_fake_azure_speech_module(context)
    behaviors = _ensure_fake_azure_recognition_behaviors(context)
    behaviors[filename] = _FakeAzureRecognitionBehavior(
        transcript="",
        reason="Canceled",
        cancellation_reason=reason,
        error_details=f"Cancellation error: {reason}"
    )


@given("the Azure Speech dependency is unavailable")
def step_azure_speech_dependency_unavailable(context) -> None:
    _install_azure_speech_unavailable_module(context)


@given("an Azure Speech API key is configured for this scenario")
def step_azure_speech_api_key_configured(context) -> None:
    api_key = os.environ.get("AZURE_SPEECH_KEY")
    if not api_key:
        api_key = "test-azure-key"
    extra_env = getattr(context, "extra_env", None)
    if extra_env is None:
        extra_env = {}
        context.extra_env = extra_env
    extra_env["AZURE_SPEECH_KEY"] = api_key


@then('the Azure Speech recognizer used region "{region}"')
def step_azure_speech_used_region(context, region: str) -> None:
    _ = context
    speechsdk_module = sys.modules.get("azure.cognitiveservices.speech")
    assert speechsdk_module is not None
    actual = getattr(speechsdk_module, "last_region", None)
    assert actual == region


@then("the Azure Speech recognizer enabled dictation")
def step_azure_speech_enabled_dictation(context) -> None:
    _ = context
    speechsdk_module = sys.modules.get("azure.cognitiveservices.speech")
    assert speechsdk_module is not None
    enabled = getattr(speechsdk_module, "last_dictation_enabled", False)
    assert enabled is True


@then('the Azure Speech recognizer used profanity option "{option}"')
def step_azure_speech_used_profanity_option(context, option: str) -> None:
    _ = context
    speechsdk_module = sys.modules.get("azure.cognitiveservices.speech")
    assert speechsdk_module is not None
    actual = getattr(speechsdk_module, "last_profanity_option", None)
    assert actual == option, f"Expected profanity option {option!r}, but got {actual!r}"



from __future__ import annotations

import re
from typing import Optional

from behave import given, then, when

from biblicus.ai.models import AiProvider, LlmClientConfig
from biblicus.text.link import (
    _apply_link_replace,
    _build_retry_message,
    _validate_link_markup,
    _validate_preserved_text,
    apply_text_link,
)
from biblicus.text.models import TextLinkRequest
from biblicus.text.tool_loop import ToolLoopResult


def _link_system_prompt_template() -> str:
    return (
        "You are a virtual file editor. Use the available tools to edit the text.\n"
        "Interpret the user's request as: wrap the requested text with "
        '<span ATTRIBUTE="VALUE">...</span> in-place in the current text.\n'
        "Each span must include exactly one attribute: id for first mentions and ref for repeats.\n"
        "Id values must start with '{{ id_prefix }}'.\n\n"
        "Use the str_replace tool to insert span tags and the done tool when finished.\n"
        "For long spans, insert the opening and closing tags using separate str_replace calls. "
        "For short spans (a few words), it is acceptable to insert both tags in one call.\n"
        "When finished, call done. Do NOT return JSON in the assistant message.\n\n"
        "Rules:\n"
        "- Use str_replace only.\n"
        "- old_str must match exactly once in the current text.\n"
        "- When choosing old_str, copy the exact substring (including punctuation/case) from the current text.\n"
        "- old_str and new_str must be non-empty strings.\n"
        "- new_str must be identical to old_str with only <span ...> and/or </span> inserted.\n"
        "- Do not include <span or </span> inside old_str.\n"
        "- Do not insert nested spans.\n"
        "- If a tool call fails due to non-unique old_str, retry with a longer unique old_str.\n"
        "- If a tool call fails, read the error and keep editing. Do not call done until spans are inserted.\n"
        "- Do not delete, reorder, or paraphrase text.\n\n"
        "Current text:\n---\n{text}\n---\n"
    )


def _build_link_request(
    *,
    text: str,
    prompt_template: str,
    id_prefix: Optional[str] = None,
    max_rounds: int = 6,
    max_edits_per_round: int = 10,
    api_key: Optional[str] = "test-openai-key",
    timeout_seconds: Optional[float] = None,
) -> TextLinkRequest:
    return TextLinkRequest(
        text=text,
        client=LlmClientConfig(
            provider=AiProvider.OPENAI,
            model="gpt-4o-mini",
            api_key=api_key,
            response_format="json_object",
            timeout_seconds=timeout_seconds,
        ),
        prompt_template=prompt_template,
        system_prompt=_link_system_prompt_template(),
        id_prefix=id_prefix or "link_",
        max_rounds=max_rounds,
        max_edits_per_round=max_edits_per_round,
    )


@when('I apply text link to text "{text}"')
def step_apply_text_link(context, text: str) -> None:
    request = _build_link_request(text=text, prompt_template="Return the requested text.")
    context.text_link_result = apply_text_link(request)


@when('I apply text link to text "{text}" with prompt template:')
def step_apply_text_link_with_prompt(context, text: str) -> None:
    prompt_template = str(getattr(context, "text", "") or "")
    request = _build_link_request(
        text=text,
        prompt_template=prompt_template,
        api_key=None,
        timeout_seconds=300.0,
    )
    context.text_link_result = apply_text_link(request)


@when('I apply text link to text "{text}" with prompt template and default system prompt:')
def step_apply_text_link_with_prompt_default_system(context, text: str) -> None:
    prompt_template = str(getattr(context, "text", "") or "")
    request = TextLinkRequest(
        text=text,
        client=LlmClientConfig(
            provider=AiProvider.OPENAI,
            model="gpt-4o-mini",
            api_key=None,
            response_format="json_object",
            timeout_seconds=300.0,
        ),
        prompt_template=prompt_template,
        id_prefix="link_",
        max_rounds=10,
        max_edits_per_round=200,
    )
    context.text_link_result = apply_text_link(request)


@when('I attempt to apply text link to text "{text}"')
def step_attempt_text_link(context, text: str) -> None:
    request = _build_link_request(
        text=text,
        prompt_template="Return the requested text.",
        max_rounds=1,
        max_edits_per_round=5,
    )
    try:
        context.text_link_result = apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)


@then("the text link has {count:d} spans")
def step_text_link_span_count(context, count: int) -> None:
    result = context.text_link_result
    assert len(result.spans) == count


@then("the text link has at least {count:d} spans")
def step_text_link_span_count_at_least(context, count: int) -> None:
    result = context.text_link_result
    assert len(result.spans) >= count


@then('the text link span {index:d} text equals "{expected}"')
def step_text_link_span_text(context, index: int, expected: str) -> None:
    result = context.text_link_result
    assert result.spans[index - 1].text == expected


@then('the text link span {index:d} has attribute "{name}" equals "{value}"')
def step_text_link_span_attribute(context, index: int, name: str, value: str) -> None:
    result = context.text_link_result
    assert result.spans[index - 1].attributes.get(name) == value


@then("the text link has an id span")
def step_text_link_has_id_span(context) -> None:
    result = context.text_link_result
    assert any("id" in span.attributes for span in result.spans)


@then("the text link has a ref span")
def step_text_link_has_ref_span(context) -> None:
    result = context.text_link_result
    assert any("ref" in span.attributes for span in result.spans)


@then("the text link does not split words")
def step_text_link_no_word_splits(context) -> None:
    result = context.text_link_result
    marked = result.marked_up_text
    assert re.search(r"\w<span\b", marked) is None
    assert re.search(r"\w</span>\w", marked) is None


@then('the text link error mentions "{text}"')
def step_text_link_error_mentions(context, text: str) -> None:
    error = context.text_link_error
    assert error is not None
    assert text in error


@given("the text link retry errors are:")
def step_set_text_link_retry_errors(context) -> None:
    raw = str(getattr(context, "text", "") or "")
    errors = [line.strip() for line in raw.splitlines() if line.strip()]
    context.text_link_retry_errors = errors


@when("I build a text link retry message for markup:")
def step_build_text_link_retry_message(context) -> None:
    errors = getattr(context, "text_link_retry_errors", []) or []
    current_text = str(getattr(context, "text", "") or "")
    context.text_link_retry_message = _build_retry_message(errors, current_text, "link_")


@then('the text link retry message includes "{text}"')
def step_text_link_retry_message_includes(context, text: str) -> None:
    message = context.text_link_retry_message
    assert text in message


@then("the text link retry message includes span context")
def step_text_link_retry_message_has_context(context) -> None:
    message = context.text_link_retry_message
    assert "Relevant spans" in message


@when("I apply text link with a non-done tool loop result")
def step_apply_text_link_with_incomplete_tool_loop(context) -> None:
    request = _build_link_request(text="Acme Acme", prompt_template="Return the requested text.")
    replacement = '<span id="link_1">Acme</span> <span ref="link_1">Acme</span>'

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text=replacement, done=False, last_error=None, messages=[])

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    link_module.run_tool_loop = fake_tool_loop
    try:
        context.text_link_result = apply_text_link(request)
    finally:
        link_module.run_tool_loop = original


@then("the text link warnings include max rounds")
def step_text_link_warnings_include_max_rounds(context) -> None:
    warnings = context.text_link_result.warnings
    assert any("max rounds" in warning for warning in warnings)


@when("I attempt text link with a tool loop error and done")
def step_attempt_text_link_tool_loop_error_done(context) -> None:
    request = _build_link_request(text="Acme", prompt_template="Return the requested text.")

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text="Acme", done=True, last_error="tool loop error", messages=[])

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    link_module.run_tool_loop = fake_tool_loop
    try:
        apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)
    finally:
        link_module.run_tool_loop = original


@when("I attempt text link with no spans and done")
def step_attempt_text_link_no_spans(context) -> None:
    request = _build_link_request(text="Acme", prompt_template="Return the requested text.")

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text="Acme", done=True, last_error=None, messages=[])

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    original_confirm = link_module.request_confirmation
    link_module.run_tool_loop = fake_tool_loop
    link_module.request_confirmation = lambda **_kwargs: ToolLoopResult(
        text="Acme", done=True, last_error=None, messages=[]
    )
    try:
        context.text_link_result = apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)
    finally:
        link_module.run_tool_loop = original
        link_module.request_confirmation = original_confirm


@then('the text link warnings include "{text}"')
def step_text_link_warnings_include(context, text: str) -> None:
    warnings = context.text_link_result.warnings
    assert any(text in warning for warning in warnings)


@when("I attempt text link with invalid spans after the tool loop")
def step_attempt_text_link_invalid_spans_after_loop(context) -> None:
    request = _build_link_request(text="Acme", prompt_template="Return the requested text.")
    marked_up = '<span label="x">Acme</span>'

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text=marked_up, done=True, last_error=None, messages=[])

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    link_module.run_tool_loop = fake_tool_loop
    try:
        apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)
    finally:
        link_module.run_tool_loop = original


@when("I apply text link and recover from missing coverage in a non-done tool loop")
def step_apply_text_link_missing_coverage_recovery(context) -> None:
    request = _build_link_request(
        text="Acme launched. Acme updated.", prompt_template="Return the requested text."
    )
    marked_up = '<span id="link_1">Acme</span> launched. Acme updated.'

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(
            text=marked_up, done=False, last_error="validation error", messages=[]
        )

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    link_module.run_tool_loop = fake_tool_loop
    try:
        context.text_link_result = apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)
    finally:
        link_module.run_tool_loop = original


@when("I attempt text link where confirmation fails with a last error")
def step_attempt_text_link_confirmation_last_error(context) -> None:
    request = _build_link_request(text="Acme", prompt_template="Return the requested text.")

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text="Acme", done=True, last_error=None, messages=[])

    def fake_confirmation(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text="Acme", done=False, last_error="confirmation error", messages=[])

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    original_confirm = link_module.request_confirmation
    link_module.run_tool_loop = fake_tool_loop
    link_module.request_confirmation = fake_confirmation
    try:
        apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)
    finally:
        link_module.run_tool_loop = original
        link_module.request_confirmation = original_confirm


@when("I attempt text link where confirmation returns invalid linked spans")
def step_attempt_text_link_confirmation_invalid_spans(context) -> None:
    request = _build_link_request(text="Acme", prompt_template="Return the requested text.")
    marked_up = '<span id="bad_1">Acme</span>'

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text="Acme", done=True, last_error=None, messages=[])

    def fake_confirmation(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text=marked_up, done=True, last_error=None, messages=[])

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    original_confirm = link_module.request_confirmation
    link_module.run_tool_loop = fake_tool_loop
    link_module.request_confirmation = fake_confirmation
    try:
        apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)
    finally:
        link_module.run_tool_loop = original
        link_module.request_confirmation = original_confirm


@when("I apply text link where confirmation inserts linked spans")
def step_apply_text_link_confirmation_inserts_linked_spans(context) -> None:
    request = _build_link_request(
        text="Acme launched. Acme updated.",
        prompt_template="Return the requested text.",
    )
    marked_up = '<span id="link_1">Acme</span> launched. ' '<span ref="link_1">Acme</span> updated.'

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text=request.text, done=True, last_error=None, messages=[])

    def fake_confirmation(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text=marked_up, done=True, last_error=None, messages=[])

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    original_confirm = link_module.request_confirmation
    link_module.run_tool_loop = fake_tool_loop
    link_module.request_confirmation = fake_confirmation
    try:
        context.text_link_result = apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)
    finally:
        link_module.run_tool_loop = original
        link_module.request_confirmation = original_confirm


@when("I attempt text link where autofill returns invalid linked spans")
def step_attempt_text_link_autofill_invalid_spans(context) -> None:
    request = _build_link_request(
        text="Acme launched. Acme updated.",
        prompt_template="Return the requested text.",
    )
    marked_up = '<span id="link_1">Acme</span> launched. Acme updated.'

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text=marked_up, done=True, last_error=None, messages=[])

    def fake_autofill(marked_up_text: str, spans):  # type: ignore[no-untyped-def]
        return marked_up_text, list(spans), ["Autofilled ref spans (no-op)"]

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    original_autofill = link_module._autofill_ref_spans
    link_module.run_tool_loop = fake_tool_loop
    link_module._autofill_ref_spans = fake_autofill
    try:
        apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)
    finally:
        link_module.run_tool_loop = original
        link_module._autofill_ref_spans = original_autofill


@when("I apply text link and autofill missing ref spans")
def step_apply_text_link_autofill_missing_refs(context) -> None:
    request = _build_link_request(
        text="Acme launched. Acme updated.", prompt_template="Return the requested text."
    )
    marked_up = '<span id="link_1">Acme</span> launched. Acme updated.'

    def fake_tool_loop(**_kwargs: object) -> ToolLoopResult:
        return ToolLoopResult(text=marked_up, done=True, last_error=None, messages=[])

    from biblicus.text import link as link_module

    original = link_module.run_tool_loop
    link_module.run_tool_loop = fake_tool_loop
    try:
        context.text_link_result = apply_text_link(request)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)
    finally:
        link_module.run_tool_loop = original


@when('I attempt link replace with old_str "{old_str}" and new_str "{new_str}"')
def step_attempt_link_replace(context, old_str: str, new_str: str) -> None:
    try:
        _apply_link_replace("Acme Acme", old_str, new_str)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)


@when('I attempt link replace in text "{text}" with old_str "{old_str}" and new_str "{new_str}"')
def step_attempt_link_replace_in_text(context, text: str, old_str: str, new_str: str) -> None:
    try:
        _apply_link_replace(text, old_str, new_str)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)


@when("I attempt to validate link markup:")
def step_attempt_validate_link_markup(context) -> None:
    markup = str(getattr(context, "text", "") or "")
    errors = _validate_link_markup(markup, "link_")
    if errors:
        context.text_link_error = "; ".join(errors)
    else:
        context.text_link_error = None


@when(
    'I attempt to validate link preservation with original "{original}" and marked up "{marked_up}"'
)
def step_attempt_validate_link_preservation(context, original: str, marked_up: str) -> None:
    try:
        _validate_preserved_text(original=original, marked_up=marked_up)
        context.text_link_error = None
    except Exception as exc:  # noqa: BLE001
        context.text_link_error = str(exc)



from __future__ import annotations

import json
import os
import shutil
import stat
import subprocess
import sys
import types
from pathlib import Path

from behave import given, then, when
from pydantic import BaseModel, ConfigDict

from biblicus.graph import get_graph_extractor
from biblicus.graph.base import GraphExtractor
from biblicus.graph.models import GraphExtractionResult
from biblicus.models import CatalogItem
from biblicus.corpus import Corpus
from features.environment import run_biblicus


class _DummyGraphExtractor(GraphExtractor):
    extractor_id = "dummy"

    def validate_config(self, config: dict[str, object]) -> BaseModel:
        _ = config
        raise NotImplementedError

    def extract_graph(
        self,
        *,
        corpus: Corpus,
        item: CatalogItem,
        extracted_text: str,
        config: BaseModel,
    ) -> GraphExtractionResult:
        _ = corpus
        _ = item
        _ = extracted_text
        _ = config
        return super().extract_graph(
            corpus=corpus,
            item=item,
            extracted_text=extracted_text,
            config=config,
        )


class _DummyGraphConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")


def _corpus_path(context, name: str) -> Path:
    return (context.workdir / name).resolve()


def _table_key_value(row) -> tuple[str, str]:
    if "key" in row.headings and "value" in row.headings:
        return row["key"].strip(), row["value"].strip()
    return row[0].strip(), row[1].strip()


def _parse_json_output(standard_output: str) -> dict[str, object]:
    return json.loads(standard_output)


def _snapshot_reference_from_context(context, name: str) -> str:
    snapshots = getattr(context, "graph_snapshot_refs", {})
    reference = snapshots.get(name)
    assert isinstance(reference, str)
    return reference


def _install_fake_neo4j_module(context) -> None:
    if getattr(context, "_fake_neo4j_installed", False):
        return
    context._fake_neo4j_original_module = sys.modules.get("neo4j")
    calls = []

    class _FakeSession:
        def __init__(self, database=None):
            self.database = database

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            return False

        def run(self, query, **params):
            calls.append(("run", query, params))

        def execute_write(self, func, *args):
            tx = _FakeTx(calls)
            return func(tx, *args)

    class _FakeTx:
        def __init__(self, call_log):
            self._call_log = call_log

        def run(self, query, **params):
            self._call_log.append(("tx_run", query, params))

    class _FakeDriver:
        def __init__(self):
            self._closed = False

        def session(self, database=None):
            return _FakeSession(database=database)

        def close(self):
            self._closed = True

    class _FakeGraphDatabase:
        @staticmethod
        def driver(_uri, auth=None):
            calls.append(("driver", _uri, auth))
            return _FakeDriver()

    fake_module = types.ModuleType("neo4j")
    fake_module.GraphDatabase = _FakeGraphDatabase
    sys.modules["neo4j"] = fake_module
    context._fake_neo4j_installed = True
    context.fake_neo4j_calls = calls


def _install_fake_spacy_module(context) -> None:
    if getattr(context, "_fake_spacy_installed", False):
        return
    context._fake_spacy_original_module = sys.modules.get("spacy")

    class _FakeSpan:
        def __init__(self, text: str, label: str):
            self.text = text
            self.label_ = label

    class _FakeToken:
        def __init__(self, text: str):
            self.text = text
            self.lemma_ = text.lower()
            self.dep_ = "ROOT"
            self.head = self

    class _FakeDoc:
        def __init__(self, text: str):
            self.text = text
            self.ents = _extract_fake_ents(text)
            self._tokens = [_FakeToken(token) for token in text.split()]

        def __iter__(self):
            return iter(self._tokens)

    class _FakeNlp:
        def __init__(self, name: str):
            self.name = name

        def __call__(self, text: str):
            return _FakeDoc(text)

    def _extract_fake_ents(text: str):
        import re

        matches = re.findall(r"\b[A-Z][a-z]+(?:\s+[A-Z][a-z]+)*\b", text)
        return [_FakeSpan(match, "PERSON") for match in matches]

    def load(name: str):
        return _FakeNlp(name)

    fake_module = types.ModuleType("spacy")
    fake_module.load = load
    sys.modules["spacy"] = fake_module
    context._fake_spacy_installed = True


def _install_fake_docker(context) -> None:
    if getattr(context, "_fake_docker_installed", False):
        return
    bin_dir = context.workdir / "fake-bin"
    bin_dir.mkdir(parents=True, exist_ok=True)
    state_path = bin_dir / "docker-state.json"
    log_path = bin_dir / "docker-log.txt"
    state_path.write_text(json.dumps({"exists": False, "running": False}) + "\n")
    log_path.write_text("")

    script_path = bin_dir / "docker"
    python_path = sys.executable
    script = f"""#!/bin/sh
set -e
STATE={state_path}
LOG={log_path}
echo "$@" >> "$LOG"
cmd="$1"
shift
if [ "$cmd" = "ps" ]; then
  exists=$({python_path} - <<'PY'
import json
import os
state=json.load(open(os.environ['STATE']))
print('1' if state.get('exists') else '0')
PY
)
  running=$({python_path} - <<'PY'
import json
import os
state=json.load(open(os.environ['STATE']))
print('1' if state.get('running') else '0')
PY
)
  if echo "$@" | grep -q "status=running"; then
    if [ "$running" = "1" ]; then
      echo "biblicus-neo4j"
    fi
    exit 0
  fi
  if [ "$exists" = "1" ]; then
    echo "biblicus-neo4j"
  fi
  exit 0
fi
if [ "$cmd" = "start" ]; then
  {python_path} - <<'PY'
import json
import os
path=os.environ['STATE']
state=json.load(open(path))
state['exists']=True
state['running']=True
json.dump(state, open(path, 'w'))
PY
  exit 0
fi
if [ "$cmd" = "run" ]; then
  {python_path} - <<'PY'
import json
import os
path=os.environ['STATE']
state=json.load(open(path))
state['exists']=True
state['running']=True
json.dump(state, open(path, 'w'))
PY
  exit 0
fi
echo "unsupported docker command" >&2
exit 1
"""
    script_path.write_text(script)
    script_path.chmod(script_path.stat().st_mode | stat.S_IEXEC)

    context._fake_docker_installed = True
    context.fake_docker_state = state_path
    context.fake_docker_log = log_path
    extra_env = getattr(context, "extra_env", {})
    extra_env["PATH"] = f"{bin_dir}:{os.environ.get('PATH','')}"
    extra_env["STATE"] = str(state_path)
    context.extra_env = extra_env


def _docker_log_contains(context, text: str) -> bool:
    log_path = getattr(context, "fake_docker_log", None)
    if log_path is None:
        return False
    return text in log_path.read_text()


@when('I attempt to resolve graph extractor "{extractor_id}"')
def step_attempt_resolve_graph_extractor(context, extractor_id: str) -> None:
    try:
        get_graph_extractor(extractor_id)
        context.graph_error = None
    except KeyError as exc:
        context.graph_error = exc


@then('the graph extractor error mentions "{text}"')
def step_graph_extractor_error_mentions(context, text: str) -> None:
    error = getattr(context, "graph_error", None)
    assert error is not None
    assert text in str(error)


@when("I invoke the graph extractor base class")
def step_invoke_graph_extractor_base(context) -> None:
    extractor = _DummyGraphExtractor()
    try:
        extractor.extract_graph(
            corpus=Corpus.init(context.workdir / "corpus", force=True),
            item=CatalogItem(
                id="item",
                relpath="item.txt",
                sha256="",
                bytes=0,
                media_type="text/plain",
                tags=[],
                metadata={},
                created_at="2024-01-01T00:00:00Z",
                source_uri="file://item.txt",
            ),
            extracted_text="",
            config=_DummyGraphConfig(),
        )
        context.graph_error = None
    except NotImplementedError as exc:
        context.graph_error = exc
        context.analysis_error = exc


@when(
    'I attempt to build a "{extractor_id}" graph extraction snapshot in corpus "{corpus_name}" with extraction snapshot "{snapshot_ref}"'
)
def step_attempt_build_graph_snapshot_missing_extraction(
    context, extractor_id: str, corpus_name: str, snapshot_ref: str
) -> None:
    corpus = _corpus_path(context, corpus_name)
    args = [
        "--corpus",
        str(corpus),
        "graph",
        "extract",
        "--extractor",
        extractor_id,
        "--extraction-snapshot",
        snapshot_ref,
    ]
    context.last_result = run_biblicus(context, args, extra_env=getattr(context, "extra_env", None))


@given("a fake Neo4j driver is installed")
@when("a fake Neo4j driver is installed")
def step_install_fake_neo4j(context) -> None:
    _install_fake_neo4j_module(context)


@given("a fake Docker daemon is installed for Neo4j")
@when("a fake Docker daemon is installed for Neo4j")
def step_install_fake_docker(context) -> None:
    _install_fake_docker(context)


@given("a fake NLP model is installed")
@when("a fake NLP model is installed")
def step_install_fake_spacy(context) -> None:
    _install_fake_spacy_module(context)


@when(
    'I build a "{extractor_id}" graph extraction snapshot in corpus "{corpus_name}" using the latest extraction snapshot and config:'
)
def step_build_graph_snapshot_latest_extraction(context, extractor_id: str, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    _install_fake_neo4j_module(context)
    if not getattr(context, "_fake_docker_installed", False):
        extra_env = getattr(context, "extra_env", {})
        extra_env["BIBLICUS_NEO4J_AUTO_START"] = "false"
        context.extra_env = extra_env
    args = ["--corpus", str(corpus), "graph", "extract", "--extractor", extractor_id]
    for row in context.table:
        key, value = _table_key_value(row)
        args.extend(["--override", f"{key}={value}"])
    result = run_biblicus(context, args, extra_env=getattr(context, "extra_env", None))
    assert result.returncode == 0, result.stderr
    context.last_graph_snapshot = _parse_json_output(result.stdout)
    context.last_graph_snapshot_id = context.last_graph_snapshot.get("snapshot_id")
    context.last_graph_extractor_id = extractor_id


@when(
    'I build a "{extractor_id}" graph extraction snapshot in corpus "{corpus_name}" using the latest extraction snapshot with real Neo4j and config:'
)
def step_build_graph_snapshot_latest_extraction_real(
    context, extractor_id: str, corpus_name: str
) -> None:
    corpus = _corpus_path(context, corpus_name)
    if shutil.which("docker") is None:
        scenario = getattr(context, "scenario", None)
        if scenario is not None:
            scenario.skip("Docker is required for Neo4j integration scenarios.")
        return
    docker_info = subprocess.run(
        ["docker", "info"],
        capture_output=True,
        text=True,
        check=False,
    )
    if docker_info.returncode != 0:
        scenario = getattr(context, "scenario", None)
        if scenario is not None:
            scenario.skip("Docker daemon is required for Neo4j integration scenarios.")
        return
    extra_env = getattr(context, "extra_env", {})
    extra_env["BIBLICUS_NEO4J_AUTO_START"] = "true"
    context.extra_env = extra_env
    args = ["--corpus", str(corpus), "graph", "extract", "--extractor", extractor_id]
    for row in context.table:
        key, value = _table_key_value(row)
        args.extend(["--override", f"{key}={value}"])
    result = run_biblicus(context, args, extra_env=getattr(context, "extra_env", None))
    assert result.returncode == 0, result.stderr
    context.last_graph_snapshot = _parse_json_output(result.stdout)
    context.last_graph_snapshot_id = context.last_graph_snapshot.get("snapshot_id")
    context.last_graph_extractor_id = extractor_id


@when('I remember the last graph extraction snapshot reference as "{name}"')
def step_remember_graph_snapshot_reference(context, name: str) -> None:
    extractor_id = context.last_graph_extractor_id
    snapshot_id = context.last_graph_snapshot_id
    assert isinstance(extractor_id, str) and extractor_id
    assert isinstance(snapshot_id, str) and snapshot_id
    reference = f"{extractor_id}:{snapshot_id}"
    snapshots = getattr(context, "graph_snapshot_refs", {})
    snapshots[name] = reference
    context.graph_snapshot_refs = snapshots


@then('the last graph extraction snapshot reference equals "{name}"')
def step_last_graph_snapshot_reference_equals(context, name: str) -> None:
    reference = _snapshot_reference_from_context(context, name)
    extractor_id = context.last_graph_extractor_id
    snapshot_id = context.last_graph_snapshot_id
    combined = f"{extractor_id}:{snapshot_id}"
    assert combined == reference


@then('the last graph extraction snapshot reference does not equal "{name}"')
def step_last_graph_snapshot_reference_not_equals(context, name: str) -> None:
    reference = _snapshot_reference_from_context(context, name)
    extractor_id = context.last_graph_extractor_id
    snapshot_id = context.last_graph_snapshot_id
    combined = f"{extractor_id}:{snapshot_id}"
    assert combined != reference


@when('I list graph extraction snapshots in corpus "{corpus_name}"')
def step_list_graph_snapshots(context, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    args = ["--corpus", str(corpus), "graph", "list"]
    result = run_biblicus(context, args, extra_env=getattr(context, "extra_env", None))
    assert result.returncode == 0, result.stderr
    context.graph_snapshot_list = _parse_json_output(result.stdout)


@then('the graph extraction snapshot list includes "{name}"')
def step_graph_snapshot_list_includes(context, name: str) -> None:
    reference = _snapshot_reference_from_context(context, name)
    entries = getattr(context, "graph_snapshot_list", None)
    assert entries is not None
    references = {f"{entry['extractor_id']}:{entry['snapshot_id']}" for entry in entries}
    assert reference in references


@then("the graph extraction snapshot list is empty")
def step_graph_snapshot_list_empty(context) -> None:
    entries = getattr(context, "graph_snapshot_list", None)
    assert entries == []


@when('I show graph extraction snapshot "{name}" in corpus "{corpus_name}"')
def step_show_graph_snapshot(context, name: str, corpus_name: str) -> None:
    corpus = _corpus_path(context, corpus_name)
    reference = _snapshot_reference_from_context(context, name)
    args = ["--corpus", str(corpus), "graph", "show", "--snapshot", reference]
    result = run_biblicus(context, args, extra_env=getattr(context, "extra_env", None))
    assert result.returncode == 0, result.stderr
    context.last_graph_snapshot = _parse_json_output(result.stdout)


@then('the shown graph extraction snapshot reference equals "{name}"')
def step_graph_snapshot_reference_equals(context, name: str) -> None:
    reference = _snapshot_reference_from_context(context, name)
    snapshot = getattr(context, "last_graph_snapshot", None)
    assert snapshot is not None
    combined = f"{snapshot['configuration']['extractor_id']}:{snapshot['snapshot_id']}"
    assert combined == reference


@then('the graph extraction snapshot graph identifier starts with "{prefix}"')
def step_graph_snapshot_graph_id_prefix(context, prefix: str) -> None:
    snapshot = getattr(context, "last_graph_snapshot", None)
    assert snapshot is not None
    graph_id = snapshot.get("graph_id")
    assert isinstance(graph_id, str)
    assert graph_id.startswith(prefix)


@then("the Docker run command is invoked for Neo4j")
def step_docker_run_invoked(context) -> None:
    assert _docker_log_contains(context, "run")
